# -*- coding: utf-8 -*-
from __future__ import print_function,division,absolute_import

import logging
log = logging.getLogger(__name__)  # __name__ is "foo.bar" here

import os
import numpy as np
np.seterr(all='ignore')

try:
  import progressbar as pb
  _has_progress_bar = True
except ImportError:
  _has_progress_bar = False
  log.warn("Reccomended package: progressbar is missing")



def progressBar(N,title="Percentage"):
  if _has_progress_bar:
    widgets = [title, pb.Percentage(), ' ', pb.Bar(), ' ', pb.ETA()]
    pbar = pb.ProgressBar(widgets=widgets, maxval=N)
  else:
    class FakeBar(object):
      def __init__(self): pass
      def start(self):    pass
      def update(self,i): pass
      def finish(self):   pass
    pbar = FakeBar()
  pbar.start()
  return pbar

def saveTxt(fname,q,data,headerv=None,info=None,overwrite=True,columns=''):
  """ Write data to file 'fname' in text format.
      Inputs:
        q = x vector
        data = one or 2D array (first axis is q)
        info = dictionary (saved as '# key : value') or string
        headerv = vector to be used as header or string
  """
  if os.path.isfile(fname) and not overwrite:
    log.warn("File %s exists, returning",fname)
    return
  if isinstance(info,dict):
    header = [ "# %s : %s" %(k,str(v)) for (k,v) in info.items() ]
    header = "\n".join(header); # skip first #, will be added by np
  elif isinstance(info,str):
    header = info
  else:
    header = ""
  if isinstance(headerv,str): header += "\n%s" % headerv
  if data.ndim == 1:
    x = np.vstack( (q,data) )
  elif data.ndim == 2:
    x = np.vstack( (q,data) )
    if headerv is not None:
      headerv = np.concatenate(( (data.shape[1],),headerv))
      x = np.hstack( (headerv[:,np.newaxis],x) )
  if columns != '':
    s = "#" + " ".join( [str(c).center(12) for c in columns] )
    header = header + "\n" + s if header != '' else s
  np.savetxt(fname,x.T,fmt="%+10.5e",header=header,comments='')

def logToScreen():
  """ It allows printing to terminal on top of logfile """
  # define a Handler which writes INFO messages or higher to the sys.stderr
  console = logging.StreamHandler()
  console.setLevel(logging.INFO)
  # set a format which is simpler for console use
  formatter = logging.Formatter('%(message)s')
  # tell the handler to use this format
  console.setFormatter(formatter)
  # add the handler to the root logger (if needed)
  if len(logging.getLogger('').handlers)==1:
    logging.getLogger('').addHandler(console)





# -*- coding: utf-8 -*-
from __future__ import print_function,division,absolute_import

import logging
log = logging.getLogger(__name__)  # __name__ is "foo.bar" here

import os
import glob
import pathlib
import re
import numpy as np
from datastorage import DataStorage

def getFiles(folder,basename="*.edf*",nFiles=None):
  files = glob.glob(os.path.join(folder,basename))
  files.sort()
  if nFiles is not None: files = files[:nFiles]
  return files

def getEdfFiles(folder,**kw):
  return getFiles(folder,basename="*.edf*",**kw)

def removeExt(fname):
  """ special remove extension meant to work with compressed files.edf and .edf.gz files """
  if fname[-3:] == ".gz": fname = fname[:-3]
  return os.path.splitext(fname)[0]

def getBasename(fname):
  return os.path.basename(removeExt(fname));

def readLogFile(fname,skip_first=0,last=None,converters=None,
    output="datastorage"):
  """ read generic log file efficiently
      lines starting with "#" will be skipped
      last line starting with # will be used to find the keys
      converters is used convert a certain field. (see np.genfromtxt)
      
      output is a string, if 'datastorage' or 'dict' data is converted
      else it is left as recarray
  """
  # makes 'output' case insentive
  if isinstance(output,str): output = output.lower()

  with open(fname,"r") as f: lines = f.readlines()
  lines = [line.strip() for line in lines]

  # find last line starting with "#"
  for iline,line in enumerate(lines):
      if line.lstrip()[0] != "#": break

  # extract names (numpy can do it but gets confused with "# dd#
  # as it does not like the space ...
  names = lines[iline-1][1:].split()

  data=np.genfromtxt(fname,skip_header=iline,names=names,dtype=None,
       converters = converters, excludelist = [] )

  # skip firsts/lasts
  data = data[skip_first:last]
  
  # force previously found names, numpy changes file to file_
  names = [ name.strip("_") for name in data.dtype.names ]
  data.dtype.names = names

  # convert to string columns that can be
  dtype = data.dtype.descr
  newtype = []
  for (name,type_str) in dtype:
      name = name.strip("_"); # numpy changes file to file_
      type_str = type_str.replace("|S","<U")
      newtype.append( (name,type_str) )
  data = data.astype(newtype) 
  

  if output.lower() == "dict":
      # convert to dict
      data = dict((name,data[name]) for name in data.dtype.names )
  elif output.lower() == "datastorage":
      data = dict((name,data[name]) for name in data.dtype.names )
      data = DataStorage( data )

  return data

  



# -*- coding: utf-8 -*-
from __future__ import print_function,division,absolute_import
from . import azav
from . import utils
from . import mask
from . import cell
from . import filters
from . import id9
from . import dataReduction
from . import center
from datastorage import DataStorage, read, save
try:
  from . import peaks
except ImportError as err:
  print("Can't import submodule peaks, reason was:",e)

__version__ = "0.5.12"



import numpy as np
import matplotlib.pyplot as plt
import os
import fabio
from .mask import interpretMasks

from datastorage import DataStorage

def _calc_R(x,y, xc, yc):
  """ calculate the distance of each 2D points from the center (xc, yc) """
  return np.sqrt((x-xc)**2 + (y-yc)**2)

def _chi2(c, x, y):
  """ calculate the algebraic distance between the data points and the mean
      circle centered at c=(xc, yc) """
  Ri = _calc_R(x, y, *c)
  return Ri - Ri.mean()

def leastsq_circle(x,y):
  from scipy import optimize
  # coordinates of the barycenter
  center_estimate = np.nanmean(x), np.nanmean(y)
  center, ier = optimize.leastsq(_chi2, center_estimate, args=(x,y))
  xc, yc = center
  Ri       = _calc_R(x, y, *center)
  R        = Ri.mean()
  residu   = np.sum((Ri - R)**2)
  return xc, yc, R

def find_center_using_circle(img,X=None,Y=None,mask=None,percentile=(90,99),\
    plot=False):
    """ Find beam center position fitting a ring (usually liquid peak)

        Parameters
        ==========
        img: array or string
          image to use, if string, reads it with fabio
        X,Y: None or arrays
          position of center of pixels, if given they will have to have
          same shape as img, if None, they will be created with meshgrid
        mask: bool array
          Pixels with True are masked out in calculations
        percentile: tuple
          range of intensity to use (in percentile values)
    """
    try:
        f = fabio.open(img)
        img = f.data
        del f
    except:
        pass
    # make sure that is float (and create a local copy) 
    img = img.astype(float)
    if mask is not None and not isinstance(mask,np.ndarray):
        mask = interpretMasks(mask,img.shape)
    if mask is not None:
        img[mask] = np.nan
    zmin,zmax = np.nanpercentile(img.ravel(),percentile[:2])
    shape     = img.shape
    idx       = (img>=zmin) & (img<=zmax)
    if X is None or Y is None:
        _use_imshow = True
        X,Y = np.meshgrid ( range(shape[1]),range(shape[0]) )
    else:
        _use_imshow = False
    xfit = X[idx].ravel()
    yfit = Y[idx].ravel()
    xc,yc,R = leastsq_circle(xfit,yfit)
    if plot:
        ax = plt.gca()
        cmin = np.nanmin(img)
        if _use_imshow:
            plt.imshow(img,clim=(cmin,zmin),cmap=plt.cm.gray)
            # RGBA
            img = np.zeros( (img.shape[0],img.shape[1],4) )
            img[idx,0]=1
            img[idx,3]=1
            plt.imshow(img)
        else:
            plt.pcolormesh(X,Y,img,cmap=plt.cm.gray,vmin=cmin,vmax=zmin)
            img = np.ma.masked_array(img,idx)
            plt.pcolormesh(X,Y,img)
        circle = plt.Circle( (xc,yc),radius=R,lw=5,color='green',fill=False)
        ax.add_artist(circle)
        plt.plot(xc,yc,"o",color="green",markersize=5)
    return DataStorage( xc=xc, yc=yc, R=R, x = xfit, y = yfit )




# -*- coding: utf-8 -*-
from __future__ import print_function,division,absolute_import
import logging
log = logging.getLogger(__name__)


import numpy as np
np.seterr(all='ignore')
import inspect
import os
import collections
import glob
import pathlib
from datastorage import DataStorage
from . import utils
from . import filters
from .mask import interpretMasks
import re
import fabio
import pyFAI


g_default_extension='.h5'
try:
  import matplotlib.pyplot as plt
except ImportError:
  log.warn("Can't import matplotlib !")

def _read(fname):
  """ read data from file using fabio """
  f = fabio.open(fname)
  data = f.data
  del f; # close file
  return data

def read(fnames):
  """ read data from file(s) using fabio """
  if isinstance(fnames,str) and os.path.isfile(fnames):
    data = _read(fnames)
  else:
    # read one image to know img size
    temp = _read(fnames[0])
    shape = [len(fnames),]+list(temp.shape)
    data  = np.empty(shape)
    data[0] = temp
    for i in range(1,len(fnames)): data[i] = _read(fnames[i])
  return data
    
def ai_as_dict(ai):
  """ ai is a pyFAI azimuthal intagrator"""
  methods = dir(ai)
  methods = [m for m in methods if m.find("get_") == 0]
  names   = [m[4:] for m in methods]
  values  = [getattr(ai,m)() for m in methods]
  ret = dict( zip(names,values) )
  ret["detector"] = ai.detector.get_name()
  return ret

def ai_as_str(ai):
  """ ai is a pyFAI azimuthal intagrator"""
  s=[ "# Detector        : %s" % ai.detector.name,
      "# Pixel      [um] : %.2fx%.2f" % (ai.pixel1*1e6,ai.pixel2*1e6),
      "# Distance   [mm] : %.3f" % (ai.dist*1e3),
      "# Center     [mm] : %.3f,%.3f" % (ai.poni1*1e3,ai.poni2*1e3),
      "# Center     [px] : %.3f,%.3f" % (ai.poni1/ai.pixel1,ai.poni2/ai.pixel2),
      "# Wavelength [A]  : %.5f" % (ai.wavelength*1e10),
      "# rot[1,2,3] [rad]: %.3f,%.3f,%.3f" % (ai.rot1,ai.rot2,ai.rot3) ]
  return "\n".join(s)

def dodezinger(ai, imgs, mask = None, npt_radial = 600, method = 'csr',dezinger=50):
  """ ai is a pyFAI azimuthal intagrator 
              it can be defined with pyFAI.load(ponifile)
        dezinger: None or float (used as percentile of ai.separate)
        mask: True are points to be masked out """
  if dezinger is None or dezinger <= 0: return imgs
  if imgs.ndim == 2: imgs=imgs[np.newaxis,:]
  for iimg,img in enumerate(imgs):
    _,imgs[iimg]=ai.separate(imgs[iimg],npt_rad=npt_radial,npt_azim=512,
                 unit='q_A^-1',method=method,mask=mask,percentile=dezinger)
  return np.squeeze(imgs)

def do1d(ai, imgs, mask = None, npt_radial = 600, method = 'csr',safe=True,dark=10., polCorr = 1,dezinger=None):
    """ ai is a pyFAI azimuthal intagrator 
              it can be defined with pyFAI.load(ponifile)
        dezinger: None or float (used as percentile of ai.separate)
        mask: True are points to be masked out """
    if isinstance(ai,dict): ai = getAI(ai)
    # force float to be sure of type casting for img
    if isinstance(dark,int): dark = float(dark);
    if imgs.ndim == 2: imgs = (imgs,)
    if mask is not None and not isinstance(mask,np.ndarray):
        mask = interpretMasks(mask,imgs[0].shape)
    out_i = np.empty( ( len(imgs), npt_radial) )
    out_s = np.empty( ( len(imgs), npt_radial) )
    for _i,img in enumerate(imgs):
      if dezinger is not None and dezinger > 0:
        img=dodezinger(ai,img,npt_radial=npt_radial,mask=mask,
                     dezinger=dezinger,method=method)
      q,i, sig = ai.integrate1d(img-dark, npt_radial, mask= mask, safe = safe,\
                 unit="q_A^-1", method = method, error_model = "poisson",
                 polarization_factor = polCorr)
      out_i[_i] = i
      out_s[_i] = sig
    return q,np.squeeze(out_i),np.squeeze(out_s)

def do2d(ai, imgs, mask = None, npt_radial = 600, npt_azim=360,method = 'csr',safe=True,dark=10., polCorr = 1):
    """ ai is a pyFAI azimuthal intagrator 
              it can be defined with pyFAI.load(ponifile)
        mask: True are points to be masked out """
    # force float to be sure of type casting for img
    if isinstance(dark,int): dark = float(dark);
    if imgs.ndim == 2: imgs = (imgs,)
    if mask is not None and not isinstance(mask,np.ndarray):
        mask = interpretMasks(mask,imgs[0].shape)
    out = np.empty( ( len(imgs), npt_azim,npt_radial) )
    for _i,img in enumerate(imgs):
      i2d,q,azTheta = ai.integrate2d(img-dark, npt_radial, npt_azim=npt_azim,
                      mask= mask, safe = safe,unit="q_A^-1", method = method,
                      polarization_factor = polCorr )
      out[_i] = i2d
    return q,azTheta,np.squeeze(out)

def getAI(poni=None,folder=None,**kwargs):
  """ get AzimuthalIntegrator instance:
      → if poni is a string, it is used as filename to read.
        in this case if folder is given it is used (together with all its 
        subfolder) as search path (along with ./ and home folder)
      → kwargs if present can be used to define (or override) parameters from files,
        dist,xcen,ycen,poni1,poni2,rot1,rot2,rot3,pixel,pixel1,pixel2,
        splineFile,detector,wavelength
  """
  if isinstance(poni,dict):
    kwargs = poni
    _def_names = ["poni1","poni2","pixel1","pixel2","dist","wavelength",\
                "rot1","rot2","rot3","xcen","ycen","pixel","E"]
    kwargs = dict( (name,value) for (name,value) in poni.items() \
             if name in _def_names )
  if isinstance(poni,pyFAI.azimuthalIntegrator.AzimuthalIntegrator):
    ai = poni
  elif isinstance(poni,str):
    # look is file exists in cwd
    if os.path.isfile(poni):
      fname = poni
    # if file does not exist look for one with that name around
    else:
      # build search paths
      folders = []
      if folder is not None:
        temp = os.path.abspath(folder)
        path = pathlib.Path(temp)
        folders = [ str(path), ]
        for p in path.parents: folders.append(str(p))
      folders.append( os.curdir )
      folders.append( os.path.expanduser("~") )
      # look for file
      for path in folders:
        fname = os.path.join(path,poni)
        if os.path.isfile(fname):
          log.info("Found poni file %s",fname)
          break
        else:
          log.debug("Could not poni file %s",fname)
    ai = pyFAI.load(fname) if os.path.isfile(fname) else None
  else:
    ai = pyFAI.azimuthalIntegrator.AzimuthalIntegrator()
  if ai is not None:
    if "E" in kwargs: kwargs["wavelength"]=12.398e-10/kwargs["E"]
    for par,value in kwargs.items(): setattr(ai,par,value)
    # provide xcen and ycen for convenience (note: xcen changes poni2
    # and ycen changes poni1)
    if 'pixel' in kwargs: ai.pixel1 = kwargs['pixel']
    if 'pixel' in kwargs: ai.pixel2 = kwargs['pixel']
    if 'xcen' in kwargs: ai.poni2 = kwargs['xcen'] * ai.pixel2
    if 'ycen' in kwargs: ai.poni1 = kwargs['ycen'] * ai.pixel1
    ai.reset(); # needed in case of overridden parameters
  else:
    log.warn("Could not find poni")
  return ai

def doFolder(folder="./",files='*.edf*',nQ = 1500,force=False,mask=None,dark=10,
    qlims=None,monitor='auto',save_pyfai=False,saveChi=True,poni='pyfai.poni',
    storageFile='auto',save=True,logDict=None,dezinger=None,skip_first=0,
    last=None):
  """ calculate 1D curves from files in folder

      Parameters
      ----------
      folder : str
          folder to work on
      files : str
          regular expression to look for ccd images (use edf* for including
          gzipped giles)
      nQ : int
          number of Q-points (equispaced)
      monitor : array or (qmin,qmax)
          normalization array (or list for q range normalization)
      force : True|False
          if True, redo from beginning even if previous data are found
          if False, do only new files
      mask : can be a list of [filenames|array of booleans|mask string]
          pixels that are True are dis-regarded
      saveChi : True|False
          if False, chi files (text based for each image) are not saved
      dezinger : None or 0<float<100
          use pyfai function 'separate' to remove zingers. The value is the 
          percentile used to find the liquicd baseline, 50 (i.e. median value)
          if a good approximation. Dezinger takes ~200ms per 4M pixel image.
          Needs good center and mask
      logDict : None or dictionary(-like)
          each key is a field. if given it has to have 'file' key
      poni : informationation necessary to build an AzimuthalIntegrator:
          → an AzimuthalIntegrator instance
          → a filename that will be look for in
               1 'folder' first
               2 in ../folder
               3 in ../../folder
               ....
               n-1 in pwd
               n   in homefolder
          → a dictionary (use to bootstrap an AzimuthalIntegrator using 
              AzimuthalIntegrator(**poni)
      save_pyfai : True|False
          if True, it stores all pyfai's internal arrays (~110 MB)
      skip_first : int
          skip the first images (the first one is sometime not ideal)
      last : int
          skip evey image after 'last'
 """

  func = inspect.currentframe()
  args = inspect.getargvalues(func)
  files_reg = files
  # store argument for saving ..
  args = dict( [(arg,args.locals[arg]) for arg in args.args] )

  folder = folder.replace("//","/").rstrip("/")

  # can't store aritrary objects
  if isinstance(args['poni'],pyFAI.AzimuthalIntegrator):
    args['poni'] = ai_as_dict(args['poni'])

  if storageFile == 'auto':
    fname = "pyfai_1d" + g_default_extension
    if not os.path.isdir(folder): 
      # do not overide folder, it might be useful
      storageFile = os.path.join( ".",fname)
    else:
      storageFile = os.path.join(folder,fname)

  if os.path.isfile(storageFile) and not force:
    saved = DataStorage(storageFile)
    log.info("Found %d images in storage file"%saved.data.shape[0])
    ai = getAI(poni,folder)
    # consistency check (saved images done with same parameters ?)
    if ai is not None:
      keys_to_compare = "nQ mask dark poni logDict dezinger skip_first last"
      keys_to_compare = keys_to_compare.split()
      # recursively transform in plain dict
      saved_args = saved.args.toDict()
      saved_args = dict( [(k,saved_args[k]) for k in keys_to_compare] )
      now_args   = dict( [(k,args[k])       for k in keys_to_compare] )
      if (saved.pyfai_info != ai_as_str(ai)) or  \
          np.any( saved.mask != interpretMasks(mask,saved.mask.shape))  or \
          (saved_args != now_args) :
        log.warn("Found inconsistency between curves already saved and new ones")
        log.warn("Redoing saved ones with new parameters")
        if (saved.pyfai_info != ai_as_str(ai)):
            log.warn("Saved pyfai parameters %s"%saved.pyfai_info)
            log.warn("New pyfai parameters %s"%ai_as_str(ai))
        if np.any(saved.mask != interpretMasks(mask,saved.mask.shape)):
            log.warn("Masks are different")
            log.warn("Old mask",saved.mask)
            log.warn("New mask",interpretMasks(mask,saved.mask.shape))
        if saved_args != now_args:
            for k in now_args.keys():
                if saved_args[k] != now_args[k]:
                    log.warn("Parameter '%s' changed from %s to %s"%\
                            (k,saved_args[k],now_args[k]))
        args['force'] = True
        saved = doFolder(**args)
  else:
    saved = None

  files = utils.getFiles(folder,files)
  if logDict is not None:
    files = [f for f in files if utils.getBasename(f) in logDict['file'] ]

    # sometime one deletes images but not corresponding lines in logfiles...
    if len(files)<len(logDict['file']):
      basenames = np.asarray( [ utils.getBasename(file) for file in files] )
      idx_to_keep = np.asarray([f in basenames for f in logDict['file']] )
      for key in logDict.keys(): logDict[key] = logDict[key][idx_to_keep]
      log.warn("More files in log than actual images, truncating loginfo")
      
  files = files[skip_first:last]


  if saved is not None:
    files = [f for f in files if utils.getBasename(f) not in saved["files"]]
  log.info("Will do azimuthal integration for %d files"%(len(files)))

  files = np.asarray(files)
  basenames = np.asarray( [ utils.getBasename(file) for file in files] )

  if len(files) > 0:
    # which poni file to use:
    ai = getAI(poni,folder)
    _msg = "could not interpret poni info or find poni file"
    if ai is None: raise ValueError(_msg)

    shape = read(files[0]).shape
    mask = interpretMasks(mask,shape)
      
    data   = np.empty( (len(files),nQ) )
    err    = np.empty( (len(files),nQ) )
    pbar = utils.progressBar(len(files))
    for ifname,fname in enumerate(files):
      img = read(fname)
      q,i,e = do1d(ai,img,mask=mask,npt_radial=nQ,dark=dark,dezinger=dezinger)
      data[ifname] = i
      err[ifname]  = e
      if saveChi:
        chi_fname = utils.removeExt(fname) + ".chi"
        utils.saveTxt(chi_fname,q,np.vstack((i,e)),info=ai_as_str(ai),overwrite=True)
      pbar.update(ifname+1)
    pbar.finish()
    if saved is not None:
      files = np.concatenate( (saved.orig.files  ,basenames ) )
      data  = np.concatenate( (saved.orig.data ,data  ) )
      err   = np.concatenate( (saved.orig.err  ,err   ) )
    else:
      files = basenames
    twotheta_rad = utils.qToTwoTheta(q,wavelength=ai.wavelength)
    twotheta_deg = utils.qToTwoTheta(q,wavelength=ai.wavelength,asDeg=True)
    orig = dict(data=data.copy(),err=err.copy(),q=q.copy(),
           twotheta_deg=twotheta_deg,twotheta_rad=twotheta_rad,files=files)
    ret = dict(folder=folder,files=files,orig = orig,pyfai=ai_as_dict(ai),
          pyfai_info=ai_as_str(ai),mask=mask,args=args)
    if not save_pyfai:
      ret['pyfai']['chia'] = None
      ret['pyfai']['dssa'] = None
      ret['pyfai']['q']    = None
      ret['pyfai']['ttha'] = None
 
    ret = DataStorage(ret)

  else:
    ret = saved

  if ret is None: return None

  if qlims is not None:
    idx = (ret.orig.q>=qlims[0]) & (ret.orig.q<=qlims[1])
  else:
    idx = np.ones_like(ret.orig.q,dtype=bool)

  ret.data = ret.orig.data[:,idx]
  ret.err  = ret.orig.err[:,idx]
  ret.q    = ret.orig.q[idx]
  ret.twotheta_rad = ret.orig.twotheta_rad
  ret.twotheta_deg = ret.orig.twotheta_deg

  if monitor == 'auto':
    monitor = ret.data.mean(1)
  elif isinstance(monitor,(tuple,list)):
    idx_norm = (ret.q >= monitor[0]) & (ret.q <= monitor[1])
    monitor = ret.data[:,idx_norm].mean(1)
  ret["data_norm"] = ret.data/monitor[:,np.newaxis]
  ret["err_norm"] = ret.err/monitor[:,np.newaxis]
  ret["monitor"] = monitor[:,np.newaxis]

  # add info from logDict if provided
  if logDict is not None: ret['log']=logDict
    # sometime saving is not necessary (if one has to do it after subtracting background
  if storageFile is not None and save: ret.save(storageFile)


  return ret


def removeBackground(data,qlims=(0,10),max_iter=30,background_regions=[],force=False,
      storageFile=None,save=True,**removeBkg):
  """ similar function to the zray.utils one, this works on dataset created by
      doFolder """
  idx = utils.findSlice(data.orig.q,qlims)
  # see if there are some to do ...
  if force:
    idx_start = 0
  else:
    idx_start = len(data.data)
  if idx_start < len(data.orig.data):
    _q,_data = utils.removeBackground(data.orig.q[idx],data.orig.data[:,idx],
               max_iter=max_iter,background_regions=background_regions,**removeBkg)
    data.q = _q
    data.data  = np.concatenate( (data.data,_data  ) )
    data.err   = np.concatenate( (data.err ,data.err[idx_start,idx]   ) )
  if save: data.save(storageFile); # if None uses .filename
  return data



def _calc_R(x,y, xc, yc):
  """ calculate the distance of each 2D points from the center (xc, yc) """
  return np.sqrt((x-xc)**2 + (y-yc)**2)

def _chi2(c, x, y):
  """ calculate the algebraic distance between the data points and the mean
      circle centered at c=(xc, yc) """
  Ri = _calc_R(x, y, *c)
  return Ri - Ri.mean()

def leastsq_circle(x,y):
  from scipy import optimize
  # coordinates of the barycenter
  center_estimate = np.nanmean(x), np.nanmean(y)
  center, ier = optimize.leastsq(_chi2, center_estimate, args=(x,y))
  xc, yc = center
  Ri       = _calc_R(x, y, *center)
  R        = Ri.mean()
  residu   = np.sum((Ri - R)**2)
  return xc, yc, R

def find_center(img,psize=100e-6,dist=0.1,wavelength=0.8e-10,center=None,reference=None,**kwargs):
  """ center is the initial centr (can be None)
      reference is a reference position to be plot in 2D plots """
  plt.ion()
  kw = dict( pixel1 = psize, pixel2 = psize, dist = dist,wavelength=wavelength )
  kw.update(kwargs)
  ai =  pyFAI.azimuthalIntegrator.AzimuthalIntegrator(**kw)
  fig_img,ax_img = plt.subplots(1,1)
  fig_pyfai,ax_pyfai = plt.subplots(1,1)
  fig_pyfai = plt.figure(2)
  temp= ax_img.imshow(img)
  plt.sca(ax_img); # set figure to use for mouse interaction
  temp.set_clim( *np.percentile(img,(2,95) ) )
  ans = ""
  print("Enter 'end' when done")
  while ans != "end":
    if center is None:
      print("Click on beam center:")
      plt.sca(ax_img); # set figure to use for mouse interaction
      center = plt.ginput()[0]
    print("Selected center:",center)
    ai.set_poni1(center[1]*psize)
    ai.set_poni2(center[0]*psize)
    q,az,i = do2d(ai,img)
    mesh = ax_pyfai.pcolormesh(q,az,i)
    mesh.set_clim( *np.percentile(i,(2,95) ) )
    ax_pyfai.set_title(str(center))
    if reference is not None: ax_pyfai.axvline(reference)
    plt.pause(0.01)
    plt.draw()
    plt.draw_all()
    ans=input("Enter to continue with clinking or enter xc,yc values ")
    if ans == '':
      center = None
    else:
      try:
        center = list(map(float,ans.split(",")))
      except Exception as e:
        center = None
    if center == []: center = None
  print("Final values: (in pixels) %.3f %.3f"%(center[0],center[1]))
  return ai

def average(fileOrFolder,delays=slice(None),scale=1,norm=None,returnAll=False,plot=False,
  showTrend=False):
  data = DataStorage(fileOrFolder)
  if isinstance(delays,slice):
    idx = np.arange(data.delays.shape[0])[delays]
  elif isinstance(delays,(int,float)):
    idx = data.delays == float(delays)
  else:
    idx = data.delays < 0
  if idx.sum() == 0:
    print("No data with the current filter")
    return None
  i   = data.data[idx]
  q   = data.q
  if isinstance(norm,(tuple,list)):
    idx  = ( q>norm[0] ) & (q<norm[1])
    norm = np.nanmean(i[:,idx],axis=1)
    i = i/norm[:,np.newaxis]
  if isinstance(norm,np.ndarray):
    i = i/norm[:,np.newaxis]
  title = "%s %s" % (fileOrFolder,str(delays))
  utils.plotdata(q,i*scale,showTrend=showTrend,plot=plot,title=title)
  if returnAll:
    return q,i.mean(axis=0)*scale,i
  else:
    return q,i.mean(axis=0)*scale

#### Utilities for chi files ####
def chiRead(fname,scale=1):
  q,i = np.loadtxt(fname,unpack=True,usecols=(0,1))
  return q,i*scale

def chiPlot(fname,useTheta=False,E=12.4):
  q,i = chiRead(fname)
  lam = 12.4/E
  theta = 2*180/3.14*np.arcsin(q*lam/4/3.14)
  x = theta if useTheta else q
  plt.plot(x,i,label=fname)



def chiAverage(folder,basename="",scale=1,norm=None,returnAll=False,plot=False,showTrend=False,clim='auto'):
  files = glob.glob(os.path.joint(folder,"%s*chi"%basename))
  files.sort()
  print(files)
  if len(files) == 0:
    print("No file found (basename %s)" % basename)
    return None
  q,_ = chiRead(files[0])
  i   = np.asarray( [ chiRead(f)[1] for f in files ] )
  if isinstance(norm,(tuple,list)):
    idx  = ( q>norm[0] ) & (q<norm[1])
    norm = np.nanmean(i[:,idx],axis=1)
    i = i/norm[:,np.newaxis]
  elif isinstance(norm,np.ndarray):
    i = i/norm[:,np.newaxis]
  title = "%s %s" % (folder,basename)
  utils.plotdata(q,i,plot=plot,showTrend=showTrend,title=title,clim=clim)
  if (showTrend and plot): plt.subplot(1,2,1)
  if showTrend:
    plt.pcolormesh(np.arange(i.shape[0]),q,i.T)
    plt.xlabel("image number, 0 being older")
    plt.ylabel(r"q ($\AA^{-1}$)")
  if (showTrend and plot): plt.subplot(1,2,2)
  if plot:
    plt.plot(q,i.mean(axis=0)*scale)
  if (plot or showTrend):
    plt.title(folder+"/"+basename)
  if returnAll:
    return q,i.mean(axis=0)*scale,i
  else:
    return q,i.mean(axis=0)*scale



# -*- coding: utf-8 -*-
from __future__ import print_function,division,absolute_import

import logging
log = logging.getLogger(__name__)  # __name__ is "foo.bar" here

import re

_timeInStr_regex = re.compile("_(-?\d+\.?\d*(?:ps|ns|us|ms)?)")

def getDelayFromString(string) :
  match = _timeInStr_regex.search(string)
  return match and match.group(1) or None

_time_regex = re.compile("(-?\d+\.?\d*)((?:s|fs|ms|ns|ps|us)?)")

def strToTime(delay) :
  if isinstance(delay,bytes): delay = delay.decode('ascii')
  _time2value = dict( fs = 1e-15, ps = 1e-12, ns = 1e-9, us = 1e-6, ms = 1e-3, s = 1)

  match = _time_regex.search(delay)
  if match:
    n,t = float(match.group(1)),match.group(2)
    value = _time2value.get(t,1)
    return n*value
  else:
    return None

def timeToStr(delay,fmt="%+.0f"):
  a_delay = abs(delay)
  if a_delay >= 1:
    ret = fmt % delay + "s"
  elif 1e-3 <= a_delay < 1: 
    ret = fmt % (delay*1e3) + "ms"
  elif 1e-6 <= a_delay < 1e-3: 
    ret = fmt % (delay*1e6) + "us"
  elif 1e-9 <= a_delay < 1e-6: 
    ret = fmt % (delay*1e9) + "ns"
  elif 1e-12 <= a_delay < 1e-9: 
    ret = fmt % (delay*1e12) + "ps"
  elif 1e-15 <= a_delay < 1e-12: 
    ret = fmt % (delay*1e12) + "fs"
  elif 1e-18 <= a_delay < 1e-15: 
    ret = fmt % (delay*1e12) + "as"
  else:
    ret = str(delay) +"s"
  return ret



# -*- coding: utf-8 -*-
from __future__ import print_function,division,absolute_import

import logging
log = logging.getLogger(__name__)  # __name__ is "foo.bar" here

import numpy as np
np.seterr(all='ignore')

def volumeFraction(concentration=1,molWeight=17,density=1.347):
  """ molWeight is kDa 
      concentration in mM
      density g/ml
  """
  concentration_mg_ml = concentration*molWeight
  volume_fraction    = concentration_mg_ml/density/1e3
  return volume_fraction

def molecularMass_from_I0(I0,c,density=1.347):
    """ calculate the molecular mass in Da

        based upon:
        D Orthaber, A Bergmann and O Glatter
        J. Appl. Cryst. (2000). 33, 218-225


        Paramters
        ---------
        I0 : float
            extrapolated value (in cm-1)
        c : float
            concentration (in g/ml)
        density : float
            protein density (in g/ml)
    """
    delta_rho = 2.67e10; # [cm-2]
    delta_rho_M = delta_rho/density
    Navo = 6e23
    M = I0*Navo / (c*delta_rho_M**2)
    return M


def radToQ(theta,**kw):
  """ theta is the scattering angle (theta_out - theta_in);
      kw should be have either E or wavelength
      it returns the scattering vector in the units of wavelength """
  # Energy or wavelength should be in kw
  assert "E" in kw or "wavelength" in kw,\
    "need wavelength or E to convert rad to Q"
  # but not both
  assert not ("E" in kw and "wavelength" in kw),\
    "conflicting arguments (E and wavelength)"
  if "E" in kw: kw["wavelength"] = 12.398/kw["E"]
  return 4*np.pi/kw["wavelength"]*np.sin(theta/2)

def degToQ(theta,**kw):
  theta = theta/180.*np.pi
  return radToQ(theta,**kw)
degToQ.__doc__ = radToQ.__doc__

def qToTwoTheta(q,asDeg=False,**kw):
  """ Return scattering angle from q (given E or wavelength) """
  # Energy or wavelength should be in kw
  assert "E" in kw or "wavelength" in kw,\
    "need wavelength or E to convert rad to Q"
  # but not both
  assert not ("E" in kw and "wavelength" in kw),\
    "conflicting arguments (E and wavelength)"
  if "E" in kw: kw["wavelength"] = 12.398/kw["E"]
  theta = 2*np.arcsin(q*kw["wavelength"]/4/np.pi)
  if asDeg: theta = np.rad2deg(theta)
  return theta



# -*- coding: utf-8 -*-
""" 
  module that contains filters and outliers removal procedures 
  most of them return the data array and a dictionary with additional info
  (parameters, statistics, etc)
"""
from __future__ import print_function,division,absolute_import
from . import utils
import copy
import logging
import statsmodels.robust
log = logging.getLogger(__name__)  # __name__ is "foo.bar" here

import numpy as np
np.seterr(all='ignore')

def applyFilter(data,boolArray):
  for key in data.keys():
    if isinstance(data[key],np.ndarray) and \
       (data[key].shape[0]==boolArray.shape[0]):
      data[key] = data[key][boolArray]
    elif isinstance(data[key],dict) and key != 'orig':
      data[key]=applyFilter(data[key],boolArray)
  return data

def applyFilters(data,funcForAveraging=np.nanmean):
  # make copy in this way tr1 = trx.filters.applyFilters(tr) does not modity tr
  data = copy.deepcopy(data)
  if not "filters" in data: return data
  if not "unfiltered" in data: data.unfiltered = \
      dict( diffs_in_scan = data.diffs_in_scan,
            chi2_0=data.chi2_0,
            diff=data.diffs )
  data.diffs_in_scan = data.unfiltered.diffs_in_scan
  filters = data.filters.keys()
  for filt_name in filters:
    filt = data.filters[filt_name]
    # understand what kind of filter (q-by-q or for every image)
    if filt[0].ndim == 1:
      for nscan in range(len(data.diffs_in_scan)):
        data.diffs_in_scan[nscan] = data.diffs_in_scan[nscan][~filt[nscan]]
        data.diffs[nscan] = funcForAveraging( data.diffs_in_scan[nscan],axis=0)
    elif filt[0].ndim == 2: # q-by-q kind of filter
      for nscan in range(len(data.diffs_in_scan)):
        data.diffs_in_scan[nscan][~filt[nscan]] = np.nan
        data.diffs[nscan] = funcForAveraging( data.diffs_in_scan[nscan],axis=0)
  data.diffs_plus_ref = data.diffs+data.ref_average
  return data


def removeZingers(curves,errs=None,norm='auto',threshold=10,useDerivative=False):
  """ curves will be normalized internally 
      if errs is None, calculate mad based noise 
      useDerivative for data with trends ..
  """
  # normalize
  if norm == 'auto':
    norm = np.nanmean(curves,axis=1)
    norm = utils.reshapeToBroadcast(norm,curves)

  if useDerivative:
    data = np.gradient(curves/norn,axis=0)
  else:
    data = curves/norm

  median = np.median(data,axis=0)

  # calculate or normalize error
  if errs is None:
    errs   = statsmodels.robust.mad(data,axis=0)
  else:
    errs   = errs/norm

  diff   = np.abs(data-median)/errs
  idx    = diff > threshold
  log.debug("Removed %d zingers from %d curves"%(idx.sum(),len(curves)))
  print("Removed %d zingers from %d curves"%(idx.sum(),len(curves)))
  if idx.sum()>0:
    curves[idx]=np.nan
    #curves = np.ma.MaskedArray(data=curves,mask=idx)
  return curves

def filterOutlier(curves,errs=None,norm=None,threshold=10):
  # normalize
  if norm == 'auto':
    norm = np.nanmean(curves,axis=1)
    norm = utils.reshapeToBroadcast(n,curves)
  elif norm is None:
    norm = 1

  curves = curves/norm
  if errs is None:
    errs   = statsmodels.robust.mad(curves,axis=0)
  else:
    errs   = errs/norm

  median = np.median(curves)
  diff   = np.abs(curves-median)/errs
  chi2   = np.sum(diff**2)/len(curves)
  idx    = chi2 < threshold
  return curves[idx]

def chi2Filter(data,threshold='auto'):
  """ Contrary to removeZingers, this removes entire curves """
  if threshold == "auto":
    threshold=np.percentile(np.concatenate(data.chi2_0),95)
  idx_mask = []
  for iscan in range(len(data.diffs_in_scan)):
    idx = data.chi2_0[iscan] > threshold
    # expand along other axis (q ...)
    #idx = utils.reshapeToBroadcast(idx,data.diffsInScanPoint[iscan])
    idx_mask.append(idx)
    log.info("Chi2 mask, scanpoint: %s, curves filtereout out %d/%d (%.2f%%)"%\
              (data.scan[iscan],idx.sum(),len(idx),idx.sum()/len(idx)*100) )

  if "filters" not in data: data.filters = dict()
  if "filters_pars" not in data: data.filters_pars = dict()
  data.filters.chi2 = idx_mask
  data.filters_pars.chi2_threshold = threshold
  return data




# -*- coding: utf-8 -*-
from __future__ import print_function,division,absolute_import

import logging
log = logging.getLogger(__name__)  # __name__ is "foo.bar" here

import numpy as np
np.seterr(all='ignore')

def findSlice(array,lims):
  start = np.ravel(np.argwhere(array>lims[0]))[0]
  stop  = np.ravel(np.argwhere(array<lims[1]))[-1]
  return slice(int(start),int(stop))

def approx(values,approx_values):
    """ returns array where every value is replaced by the closest in approx_values

        This funciton is useful for rebinning; careful, can be slow with many bins...

        Example:
        -------
        approx( np.arange(0,1,0.1), [0,0.3,0.7] )
           array([ 0. ,  0. ,  0.3,  0.3,  0.3,  0.7,  0.7,  0.7,  0.7,  0.7])

    """
    # make sure they are arrays
    values = np.asarray(values)
    approx_values = np.asarray(approx_values)
    # create outter difference
    diff = np.abs(values[:,np.newaxis] - approx_values)
    args = np.argmin(diff,axis=1)
    values = approx_values[args]
    #values = np.asarray( [ approx_values[np.argmin(np.abs(v-approx_values))] for v in values] )
    return values


def rebin(values,bins):
    """ returns array where every value is replaced by the closest in approx_values

        This funciton is useful for rebinning

        Example:
        -------
        approx( np.arange(0,1,0.1), [0,0.3,0.7] )
           array([ 0. ,  0. ,  0.3,  0.3,  0.3,  0.7,  0.7,  0.7,  0.7,  0.7])

    """
    # make sure they are arrays
    bins = np.asarray(bins)
    idx = np.digitize(values,bins)
    idx[idx > bins.shape[0]-1] = bins.shape[0]-1
    return (bins[idx]+bins[idx-1])/2

def reshapeToBroadcast(what,ref):
  """ expand the 1d array 'what' to allow broadbasting to match 
      multidimentional array 'ref'. The two arrays have to same the same 
      dimensions along the first axis
  """
  if what.shape == ref.shape: return what
  assert what.shape[0] == ref.shape[0],\
    "automatic reshaping requires same first dimention"
  shape  = [ref.shape[0],] + [1,]*(ref.ndim-1)
  return what.reshape(shape)

def removeBackground(x,data,xlims=None,max_iter=100,background_regions=[],**kw):
  from dualtree import dualtree
  if data.ndim == 1: data = data[np.newaxis,:]
  if xlims is not None:
    idx = findSlice(x,xlims)
    x = x[idx]
    data = data[:,idx].copy()
  else:
    data = data.copy(); # create local copy
  # has to be a list of lists ..
  if background_regions != [] and isinstance(background_regions[0],numbers.Real):
    background_regions = [background_regions,]
  background_regions = [findSlice(x,brange) for brange in background_regions]
  for i in range(len(data)):
    data[i] = data[i] - dualtree.baseline(data[i],max_iter=max_iter,
                        background_regions=background_regions,**kw)
  return x,np.squeeze(data)



from . import absorption
from .absorption import *

from . import misc
from .misc import *

from . import numeric
from .numeric import *

from . import files
from .files import *

from . import plot
from .plot import *

from . import scattering
from .scattering import *

from . import string
from .string import *



# -*- coding: utf-8 -*-
from __future__ import print_function,division,absolute_import

import logging
log = logging.getLogger(__name__)

import os
import collections
import numpy as np
from . import azav
from . import dataReduction
from . import utils
from . import filters
from datastorage import DataStorage

default_extension = ".h5"

def _conv(x):
    try:
        x = float(x)
    except:
        x = np.nan
    return x

def _readDiagnostic(fname,retry=3):
  ntry = 0
  while ntry<retry:
    try:
      data = np.genfromtxt(fname,usecols=(2,3),\
             dtype=None,converters={3: lambda x: _conv(x)},
             names = ['fname','delay'])
      return data
    except Exception as e:
      log.warn("Could not read diagnostic file, retrying soon,error was %s"%e)
      ntry += 1
  # it should not arrive here
  raise ValueError("Could not read diagnostic file after %d attempts"%retry)

def readDiagnostic(fname):
  """ return an ordered dict dictionary of filename; for each key a rounded
      value of delay is associated """
  if os.path.isdir(fname): fname += "/diagnostics.log"
  # try to read diagnostic couple of times
  data = _readDiagnostic(fname,retry=4)
  files = data['fname'].astype(str)
  delays = data['delay']
  # skip lines that cannot be interpreted as float (like done, etc)
  idx_ok = np.isfinite( delays )
  files = files[idx_ok]
  files = np.asarray( [utils.getBasename(f) for f in files]) 
  delays = delays[idx_ok]
  delays = np.round(delays.astype(float),12)
  return dict( file = files, scan = delays) 

def _findDark(line):
  _,value = line.split(":")
  return float(value)

def _delayToNum(delay):
  if delay.decode('ascii') == 'off':
    delay = -10
  else:
    delay=utils.strToTime(delay)
  return delay

def findLogFile(folder):
  files = utils.getFiles(folder,basename='*.log')
  files.remove(os.path.join(folder,"diagnostics.log"))
  logfile = files[0]
  if len(files)>1: log.warn("Found more than one *.log file that is not diagnostics.log: %s"%files)
  return logfile

def readLogFile(fnameOrFolder,subtractDark=False,skip_first=0,
    asDataStorage=True,last=None,srcur_min=30):
    """ read id9 style logfile; 
        last before data will be used as keys ... 
        only srcur>srcur_min will be kept
        subtractDark is not needed for data collected with waxscollect
    """
    if os.path.isdir(fnameOrFolder):
      fname = findLogFile(fnameOrFolder)
    else:
      fname = fnameOrFolder
    log.info("Reading id9 logfile:",fname)

    data = utils.files.readLogFile(fname,skip_first=skip_first,last=last,\
           output = "array")
 
    # work on darks if needed
    if subtractDark:
        ## find darks
        with open(fname,"r") as f: lines = f.readlines()
        lines = [line.strip() for line in lines]
        # look only for comment lines
        lines = [ line for line in lines if line[0] == "#" ]
        for line in lines:
            if line.find("pd1 dark/sec")>=0: darks['pd1ic'] = _findDark(line)
            if line.find("pd2 dark/sec")>=0: darks['pd2ic'] = _findDark(line)
            if line.find("pd3 dark/sec")>=0: darks['pd3ic'] = _findDark(line)

        ## subtract darks
        for diode in ['pd1ic','pd2ic','pd3ic','pd4ic']:
            if diode in darks: data[diode]=data[diode]-darks[diode]*data['timeic']

    # srcur filter
    if "currentmA" in data.dtype.names:
        idx_cur = data['currentmA'] > srcur_min
        if (idx_cur.sum() < idx_cur.shape[0]*0.5):
            log.warn("Minimum srcur filter has kept only %.1f%%"%(idx_cur.sum()/idx_cur.shape[0]*100))
            log.warn("Minimum srcur: %.2f, median(srcur): %.2f"%(srcur_min,np.nanmedian(data["currentmA"]))) 
        data = data[idx_cur]
    else:
        log.warn("Could not find currentmA in logfile, skipping filtering")

    if asDataStorage:
      data = DataStorage( dict((name,data[name]) for name in data.dtype.names ) )

    return data


def doFolder_azav(folder,nQ=1500,files='*.edf*',force=False,mask=None,
  saveChi=True,poni='pyfai.poni',storageFile='auto',dark=9.9,dezinger=None,
  qlims=None,monitor='auto',skip_first=0,last=None,srcur_min=80):
  """ very small wrapper around azav.doFolder, essentially just reading
      the id9 logfile or diagnostics.log
      monitor  : normalization vector that can be given as
               1. numpy array
               2. a list (interpreted as q-range normalization)
               3. a string to look for as key in the log, e.g.
                  monitor="pd2ic" would reult in using
                  azavStorage.log.pd2ic


 """

  try:
    loginfo = readLogFile(folder,skip_first=skip_first,last=last,
              srcur_min=srcur_min)
  except Exception as e:
    log.warn("Could not read log file, trying to read diagnostics.log")
    loginfo = readDiagnostic(folder)
  if storageFile == 'auto' : storageFile = folder +  "/" + "pyfai_1d" + default_extension

  if monitor != "auto" and isinstance(monitor,str): monitor = loginfo[monitor]

  data = azav.doFolder(folder,files=files,nQ=nQ,force=force,mask=mask,
    saveChi=saveChi,poni=poni,storageFile=storageFile,logDict=loginfo,
    dark=dark,save=False,dezinger=dezinger,qlims=qlims,monitor=monitor)
  data.save(storageFile)
  return data



def doFolder_dataRed(azavStorage,funcForAveraging=np.nanmean,
                     outStorageFile='auto',reference='min',chi2_0_max='auto',
                     saveTxt=True):
  """ azavStorage if a DataStorage instance or the filename to read 
  """

  if isinstance(azavStorage,DataStorage):
    azav = azavStorage
    folder = azavStorage.folder
  elif os.path.isfile(azavStorage):
    folder = os.path.dirname(azavStorage)
    azav = DataStorage(azavStorage)
  else:
    # assume is just a folder name
    folder = azavStorage
    azavStorage  = folder +  "/pyfai_1d" + default_extension
    azav = DataStorage(azavStorage)


  # calculate differences
  tr = dataReduction.calcTimeResolvedSignal(azav.log.delay,azav.data_norm,
          err=azav.err_norm,q=azav.q,reference=reference,
          funcForAveraging=funcForAveraging,chi2_0_max=chi2_0_max)

  tr.folder = folder
  if outStorageFile == 'auto':
    if not os.path.isdir(folder): folder = "./"
    outStorageFile = folder + "/diffs" + default_extension
  tr.filename = outStorageFile

  # save txt and npz file
  if saveTxt: dataReduction.saveTxt(folder,tr,info=azav.pyfai_info)

  tr.save(outStorageFile)

  return tr

def doFolder(folder,azav_kw = dict(), datared_kw = dict(),online=True, retryMax=20,force=False):
  import matplotlib.pyplot as plt
  if folder == "./": folder = os.path.abspath(folder)
  fig = plt.figure()
  lastNum = None
  keepGoing = True
  lines   = None
  retryNum = 0
  azav_kw['force']=force
  if online: print("Press Ctrl+C to stop")
  while keepGoing and retryNum < retryMax:
    try:
      azav = doFolder_azav(folder,**azav_kw)
      # check if there are new data
      if lastNum is None or lastNum<azav.data.shape[0]:
        tr = doFolder_dataRed(azav,**datared_kw)
        if lines is None or len(lines) != tr.diffs.shape[0]:
          lines,_ = utils.plotdiffs(tr,fig=fig,title=folder)
        else:
          utils.updateLines(lines,tr.diffs)
        plt.draw()
        lastNum = azav.data.shape[0]
        retryNum = 0
      else:
        retryNum += 1
      if online: plt.pause(30)
      if force: azav_kw['force']=False; # does not make sense to have always True ...
    except KeyboardInterrupt:
      keepGoing = False
    if not online: keepGoing = False
  return azav,tr


def readMotorDump(fnameOrFolder,asDataStorage=True,\
    default_fname="motor_position_after_data_collection.txt"):
  """ 
      Read waxecollect style motor dump
      if fnameOrFolder is a folder, default_fname is read
      if asDataStorage is False:
        return recArray with fields name,user,dial
      else: return dictory like object (each motor is a key)
  """
  if os.path.isfile(fnameOrFolder):
    fname = fnameOrFolder
  else:
    fname = os.path.join(fnameOrFolder,default_fname)
  data = np.genfromtxt(fname,names=True,dtype=("<U15",float,float))
  # remove interleaved headers
  idx_to_remove = data['name'] == 'name'
  data = data[~idx_to_remove]
#  for i in range(data.shape[0]): data['name'][i] = data['name'][i].decode('ascii')
  if asDataStorage:
    motor_pos = collections.namedtuple('motor_pos',['user','dial'])
    ret = dict()
    for imotor,motor in enumerate(data['name']):
      ret[motor] = motor_pos(dial=data['dial'][imotor],user=data['user'][imotor])
    data = DataStorage(ret)
  return data
  



# -*- coding: utf-8 -*-
from __future__ import print_function,division,absolute_import

import logging
log = logging.getLogger(__name__)  # __name__ is "foo.bar" here

import numpy as np
np.seterr(all='ignore')

def attenuation_length(compound=None, density=None, natural_density=None,energy=None, wavelength=None):
  """ extend periodictable.xsf capabilities """
  import periodictable.xsf
  if energy is not None: wavelength = periodictable.xsf.xray_wavelength(energy)
  assert wavelength is not None, "scattering calculation needs energy or wavelength"
  if (np.isscalar(wavelength)): wavelength=np.array( [wavelength] )
  n = periodictable.xsf.index_of_refraction(compound=compound,
                           density=density, natural_density=natural_density,
                           wavelength=wavelength)
  attenuation_length = (wavelength*1e-10)/ (4*np.pi*np.imag(n))
  return np.abs(attenuation_length)


def transmission(compound='Si',thickness=100e-6, att_len=None,density=None, 
    natural_density=None,energy=None, wavelength=None,angle=None):
    """ calculate transmission of a given thickness of compound.
        if att_len (in m-1) is not given, it is calculated based on compound 
        information and energy or wavelength)
        if angle is given, the thickness is divided by cos(angle)
        angle has to be in deg. angle=0 means perpendicular
    """
    if att_len is None:
        att_len = attenuation_length(compound=compound,density=density,
                  natural_density=natural_density,energy=energy,
                  wavelength=wavelength)
    # takes care if both are arrays ...
    if angle is not None:
      thickness /= np.cos(np.deg2rad(angle))
    argument = np.squeeze(np.outer(thickness,1/att_len))
    return np.exp(-argument)

def A(compound='Si',thickness=100e-6, att_len=None,density=None, 
    natural_density=None,energy=None, wavelength=None,angle=None):
    """ calculate absorption of a given thickness of compound.
        if att_len (in m-1) is not given, it is calculated based on compound 
        information and energy or wavelength)
        if angle is given, the thickness is divided by cos(angle)
        angle has to be in deg. angle=0 means perpendicular
    """
    T = transmission(compound=compound,density=density,att_len=att_len,
                  natural_density=natural_density,energy=energy,
                  thickness=thickness, wavelength=wavelength,angle=angle)
    return 1-T


def _phosphorAbsorption(twotheta,mu=17700,thickness=40e-6,energy=None):
  """ return 1-np.exp(-mu*thick/np.cos(np.deg2rad(theta)))
      - mu is the neperian absorption linear coefficient (m-1)
      - twotheta is the scattering angle (in degrees)
  """
  if mu =='auto':
    assert energy is not None, "phosphorAbsorption with automatic mu requires energy"
    att_len = attenuation_length("Ce",energy=energy,density=4.75)
  else:
    att_len = 1/mu
  return A(att_len=att_len,thickness=thickness,angle=twotheta)

def phosphorCorrection(twotheta,mu=17700,thickness=40e-6,energy=None,normalizeToZeroAngle=False):
  """ helper function to correct for angle dependent absorption of the phosphor screen for an
      x-ray detector.
      return the correction factor one has to *multiply* the data for.
      - mu is the neperian absorption linear coefficient (m-1) 
          [could be 'auto' if energy is given]
      - twotheta is the scattering angle (in degrees)
  """
  corr = 1/_phosphorAbsorption(twotheta,mu=mu,thickness=thickness,energy=energy)
  if normalizeToZeroAngle:
    corr = corr*_phosphorAbsorption(0,mu=mu,thickness=thickness,energy=energy)
  return corr

def liquidSheetAbsorptionCorrection(twotheta=0,compound="H2O",thickness=300e-6,
    density=1,att_len=None,energy=None,wavelength=None,
    normalizeToZeroAngle=False):
    """ helper function for angular dependence of sample absorption

        returns the correction factor one has to *multiply* the intensity for
        the uncorrected I(theta) is divided by the T(theta)/T(0)
        where the transmittion T(theta) is:
        1/(mu*l) cos(theta)/(1-cos(theta))*(exp(-mu*l)-exp(-mu*l/cos(theta)))
        twotheta is in deg
    """
    if att_len is None : 
        att_len = attenuation_length(compound=compound,density=density,
                  energy=energy,wavelength=wavelength)
    if np.isscalar(twotheta): twotheta = np.asarray( (twotheta,) )
    twotheta = np.deg2rad(twotheta)
    idx = (twotheta < 0.001)
    twotheta[idx] = 0.001
    cos      = np.cos(twotheta)
    t1 = transmission(att_len=att_len,thickness=thickness)
    t2 = transmission(att_len=att_len,thickness=thickness/cos)
    arg = thickness/att_len
    temp = 1/arg*cos/(1-cos)*(t1-t2)
    if normalizeToZeroAngle: temp /= transmission(att_len=att_len,thickness=thickness)
    return 1/temp


def chargeToPhoton(chargeOrCurrent,compound="Si",thickness=100e-6,energy=10,e_hole_pair=3.63):
  """
    Function to convert charge (or current to number of photons (or number 
    of photons per second)
    
    Parameters
    ----------
    
    chargeOrCurrent: float or array
    compound : str
       Used to calculate 
     
  """
  # calculate absorption
  abs_fraction = A(compound=compound,energy=energy,thickness=thickness)
  chargeOrCurrent = chargeOrCurrent/abs_fraction

  e_hole_pair_energy_keV = e_hole_pair*1e-3
  n_charge_per_photon = energy/e_hole_pair_energy_keV
  # convert to Q
  charge_per_photon = n_charge_per_photon*1.60217662e-19
  nphoton = chargeOrCurrent/charge_per_photon
 
  if len(nphoton) == 1: nphoton = float(nphoton)
  return nphoton



# -*- coding: utf-8 -*-
from __future__ import print_function,division,absolute_import
import collections
import itertools
import numpy as np
from numpy import sin,cos

class Triclinic(object):
  def __init__(self,a=1,b=1,c=1,alpha=90,beta=90,gamma=90):
    self.a = a
    self.b = b
    self.c = c
    alpha  = alpha*np.pi/180
    beta   = beta*np.pi/180
    gamma  = gamma*np.pi/180
    self.alpha = alpha
    self.beta  = beta 
    self.gamma = gamma

    self._s11 = b**2 * c**2 * sin(alpha)**2
    self._s22 = a**2 * c**2 * sin(beta)**2
    self._s33 = a**2 * b**2 * sin(gamma)**2
    self._s12 = a*b*c**2*(cos(alpha) * cos(beta) - cos(gamma))
    self._s23 = a**2*b*c*(cos(beta) * cos(gamma) - cos(alpha))
    self._s13 = a*b**2*c*(cos(gamma) * cos(alpha) - cos(beta))
    self.V    = (a*b*c)*np.sqrt(1-cos(alpha)**2 - cos(beta)**2 - cos(gamma)**2 + 2*cos(alpha)*cos(beta)*cos(gamma))

  def __call__(self,h,k,l):  return self.q(h,k,l)

  def d(self,h,k,l):
    temp = self._s11*h**2 + \
           self._s22*k**2 + \
           self._s33*l**2 + \
           2*self._s12*h*k+ \
           2*self._s23*k*l+ \
           2*self._s13*h*l
    d    = self.V/np.sqrt(temp)
    return d

  def Q(self,h,k,l):
    return 2*np.pi/self.d(h,k,l)

  def reflection_list(self,maxQ=3,lim=10):
    ret=dict()
    # prepare hkl
    i = range(-lim,lim+1)
    prod = itertools.product( i,i,i )
    hkl = np.asarray( list( itertools.product( i,i,i ) )  )
    h,k,l = hkl.T
    q = self.Q(h,k,l)

    idx = q<maxQ;
    q = q[idx]
    hkl = hkl[idx]
    q = np.round(q,12)
    qunique = np.unique(q)
    ret = []
    for qi in qunique:
      reflec = hkl[ q == qi ]
      ret.append( (qi,tuple(np.abs(reflec)[0]),len(reflec),reflec) )
    return qunique,ret
   
#    for h in range(-lim,lim+1):
#      for j in range(-lim,lim+1):


class Orthorombic(Triclinic):
  def __init__(self,a=1,b=1,c=1):
    Triclinic.__init__(self,a=a,b=b,c=c,alpha=90,beta=90,gamma=90)

class Cubic(Orthorombic):
  def __init__(self,a=1):
    Orthorombic.__init__(self,a=a,b=a,c=a)


class Monoclinic(object):
  def __init__(self,a=1,b=1,c=1,beta=90.):
    Triclinic.__init__(self,a=a,b=b,c=c,alpha=90,beta=beta,gamma=90)


def plotReflections(cell_instance,ax=None,line_kw=dict(),text_kw=dict()):
  import matplotlib.pyplot as plt
  from matplotlib import lines
  import matplotlib.transforms as transforms
  _,refl_info = cell_instance.reflection_list()
  if ax is None: ax = plt.gca()

  # the x coords of this transformation are data, and the
  # y coord are axes
  trans = transforms.blended_transform_factory(ax.transData, ax.transAxes)
  txt_kw = dict( horizontalalignment='center', rotation=45)
  txt_kw.update(**text_kw)
  for reflection in refl_info[1:]:
    q,hkl,n,_ = reflection
    line = lines.Line2D( [q,q],[1,1.1],transform=trans,**line_kw)
    line.set_clip_on(False)
    ax.add_line(line)
    ax.text(q,1.15,str(hkl),transform=trans,**txt_kw)

ti3o5_lambda = Triclinic(a = 9.83776, b = 3.78674, c = 9.97069, beta = 91.2567)
ti3o5_beta   = Triclinic(a = 9.7382 , b = 3.8005 , c = 9.4333 , beta = 91.496)
#ti3o5_beta   = Monoclinic(a = 9.7382 , b = 3.8005 , c = 9.4333 , beta = 91.496)
ti3o5_alpha  = Triclinic(a = 9.8372,  b = 3.7921,  c = 9.9717)
ti3o5_alpha1  = Orthorombic(a = 9.8372,  b = 3.7921,  c = 9.9717)

si = Cubic(a=5.431020504)



# -*- coding: utf-8 -*-
from __future__ import print_function,division,absolute_import

import logging
log = logging.getLogger(__name__)  # __name__ is "foo.bar" here

import numpy as np
np.seterr(all='ignore')
import os
import glob
import pathlib
import re
import numbers
try:
  import progressbar as pb
  _has_progress_bar = True
except ImportError:
  _has_progress_bar = False
  log.warn("Reccomended package: progressbar is missing")
from datastorage import DataStorage

try:
  import matplotlib.pyplot as plt
except ImportError:
  log.warn("Can't import matplotlib !")

_time_regex =      re.compile( "(-?\d+\.?\d*(?:ps|ns|us|ms)?)")
_timeInStr_regex = re.compile("_(-?\d+\.?\d*(?:ps|ns|us|ms)?)")

def getFiles(folder,basename="*.edf*",nFiles=None):
  files = glob.glob(os.path.join(folder,basename))
  files.sort()
  if nFiles is not None: files = files[:nFiles]
  return files

def getEdfFiles(folder,**kw):
  return getFiles(folder,basename="*.edf*",**kw)

def getDelayFromString(string) :
  match = _timeInStr_regex.search(string)
  return match and match.group(1) or None

_time_regex = re.compile("(-?\d+\.?\d*)((?:s|fs|ms|ns|ps|us)?)")

def strToTime(delay) :
  if isinstance(delay,bytes): delay = delay.decode('ascii')
  _time2value = dict( fs = 1e-15, ps = 1e-12, ns = 1e-9, us = 1e-6, ms = 1e-3, s = 1)

  match = _time_regex.search(delay)
  if match:
    n,t = float(match.group(1)),match.group(2)
    value = _time2value.get(t,1)
    return n*value
  else:
    return None

def timeToStr(delay,fmt="%+.0f"):
  a_delay = abs(delay)
  if a_delay >= 1:
    ret = fmt % delay + "s"
  elif 1e-3 <= a_delay < 1: 
    ret = fmt % (delay*1e3) + "ms"
  elif 1e-6 <= a_delay < 1e-3: 
    ret = fmt % (delay*1e6) + "us"
  elif 1e-9 <= a_delay < 1e-6: 
    ret = fmt % (delay*1e9) + "ns"
  elif 1e-12 <= a_delay < 1e-9: 
    ret = fmt % (delay*1e12) + "ps"
  elif 1e-15 <= a_delay < 1e-12: 
    ret = fmt % (delay*1e12) + "fs"
  elif 1e-18 <= a_delay < 1e-15: 
    ret = fmt % (delay*1e12) + "as"
  else:
    ret = str(delay) +"s"
  return ret

def removeExt(fname):
  """ special remove extension meant to work with compressed files.edf and .edf.gz files """
  if fname[-3:] == ".gz": fname = fname[:-3]
  return os.path.splitext(fname)[0]

def getBasename(fname):
  return os.path.basename(removeExt(fname));

def progressBar(N,title="Percentage"):
  if _has_progress_bar:
    widgets = [title, pb.Percentage(), ' ', pb.Bar(), ' ', pb.ETA()]
    pbar = pb.ProgressBar(widgets=widgets, maxval=N)
  else:
    class FakeBar(object):
      def __init__(self): pass
      def start(self):    pass
      def update(self,i): pass
      def finish(self):   pass
    pbar = FakeBar()
  pbar.start()
  return pbar


def findSlice(array,lims):
  start = np.ravel(np.argwhere(array>lims[0]))[0]
  stop  = np.ravel(np.argwhere(array<lims[1]))[-1]
  return slice(int(start),int(stop))

def removeBackground(x,data,xlims=None,max_iter=100,background_regions=[],**kw):
  from dualtree import dualtree
  if data.ndim == 1: data = data[np.newaxis,:]
  if xlims is not None:
    idx = findSlice(x,xlims)
    x = x[idx]
    data = data[:,idx].copy()
  else:
    data = data.copy(); # create local copy
  # has to be a list of lists ..
  if background_regions != [] and isinstance(background_regions[0],numbers.Real):
    background_regions = [background_regions,]
  background_regions = [findSlice(x,brange) for brange in background_regions]
  for i in range(len(data)):
    data[i] = data[i] - dualtree.baseline(data[i],max_iter=max_iter,
                        background_regions=background_regions,**kw)
  return x,np.squeeze(data)
  

def plotdata(data,x=None,plotAverage=True,showTrend=True,title=None,\
      clim='auto',fig=None):
    """ Plot 2D array to look for trends/stability or outlier

        It usually works best with normalized data

        Parameters
        ----------
        data: dict-like object or 2D array.
          if dict-like it has to have keys
          'q' used for one axis
          'data' for 2D array[numimg,nQ]
        x: array or None
          if given it has to have a length of numimg
          if None will be 0,1,2,3,...numimg and interpreted
          as image number
        plotAverage: bool
          if True, plot average
        showTrend: bool
          if True, show 2D plot (most common use)
        clim: 'auto' or tuple
          color scale to use, is 'auto' 1.5%,98.5% percentiles are used
        fig:
          matplotlib fig instance to use, if None creates a new one
        title: str
          title to use for plot (if None and data['folder'] exists
          the value is used as title
          
    """
    if isinstance(data,np.ndarray):
      q = np.arange( data.shape[-1] )
    else:
      if title is None and 'folder' in data: title = data["folder"]
      q = data["q"]
      data = data["data"]
    if not (plotAverage or showTrend): return

    if x is None: x = np.arange(data.shape[0])
    if clim == 'auto': clim = np.nanpercentile(data,(1.5,98.5))
    one_plot = showTrend or plotAverage
    two_plot = showTrend and plotAverage

    if one_plot and not two_plot:
      if fig is None:
        fig,ax = plt.subplots(1,1)
      else:
        fig.clear()
        ax = fig.axes
    if two_plot:
      if fig is None:
        fig,ax = plt.subplots(2,1,sharex=True)
      else:
        ax = fig.axes
 
    ax = np.atleast_1d(ax)
    if showTrend:
      plt.sca(ax[1])
      plt.pcolormesh(q,x,data)
      plt.ylabel("image number, 0 being older")
      plt.xlabel(r"q ($\AA^{-1}$)")
      plt.clim( *clim )
    if plotAverage:
        ax[0].plot(q,np.nanmean(data,axis=0))
    if (plotAverage or showTrend) and title is not None:
      plt.title(title)

def volumeFraction(concentration=1,molWeight=17,density=1.347):
  """ molWeight is kDa 
      concentration in mM
      density g/ml
  """
  concentration_mg_ml = concentration*molWeight
  volume_fraction    = concentration_mg_ml/density/1e3
  return volume_fraction

def molecularMass_from_I0(I0,c,density=1.347):
    """ calculate the molecular mass in Da

        based upon:
        D Orthaber, A Bergmann and O Glatter
        J. Appl. Cryst. (2000). 33, 218-225


        Paramters
        ---------
        I0 : float
            extrapolated value (in cm-1)
        c : float
            concentration (in g/ml)
        density : float
            protein density (in g/ml)
    """
    delta_rho = 2.67e10; # [cm-2]
    delta_rho_M = delta_rho/density
    Navo = 6e23
    M = I0*Navo / (c*delta_rho_M**2)
    return M

def plotdiffs(data,select=None,err=None,absSignal=None,absSignalScale=10,
           showErr=False,cmap=plt.cm.jet,fig=None,title=None,plotDiffRef=False):
    """ Plot difference data

        Parameters
        ----------
        data: dict-like object or (q,diffs,scan)tuple
          if dict-like it has to have keys
          'q' used for one axis
          'diffs' for 2D array[numscan,nQ]
          'scan'  scan points
          if present 'err' is used for errorbars
          if present 'diffs_plus_ref' is used
        select: slice
          to select time delays (for example select=slice(None,None,2))
        absSignal: array[nQ]
          if present, ass this data to the plot (could be the average)
        absSignalScale: float
          the factor to use to divide the absSignal for, it is kept as
          different parameters because the scaling it is added to the legend
        showErr: bool
          if True, plot errorbars as well
        cmap: matplotlib color map
          color map to use
        fig:
          matplotlib fig instance to use, if None creates a new one
        title: str
          title to use for plot (if None, no title is added)
        plotDiffRef:
          plot not only the diff but also the diffs_plus_ref (if present in data)
    """
    if isinstance(data,(list,tuple)):
        q,diffs,scan = args
        diffs_abs = None
    else:
        q = data["q"]
        diffs = data["diffs"]
        scan  = data["scan"]
        err   = data.get("err",None)
        diffs_abs = data.get("diffs_plus_ref",None)

    # this selection trick done in this way allows to keep the same colors when 
    # subselecting (because I do not change the size of diffs)
    if select is not None:
        indices = range(*select.indices(scan.shape[0]))
    else:
        indices = range(len(scan))
 
    if fig is None: fig = plt.figure()
 
    lines_diff = []
    lines_abs = []
    if absSignal is not None:
        line = plt.plot(q,absSignal/absSignalScale,lw=3,
                      color='k',label="absSignal/%s"%str(absSignalScale))[0]
        lines.append(line)
    for linenum,idiff in enumerate(indices):
        color = cmap(idiff/(len(diffs)-1))
        label = timeToStr(scan[idiff])
        kw = dict( color = color, label = label )
        if err is not None and showErr:
            line = plt.errorbar(q,diffs[idiff],err[idiff],**kw)[0]
            lines_diff.append(line)
        else:
            line = plt.plot(q,diffs[idiff],**kw)[0]
            lines_diff.append(line)
            if diffs_abs is not None and plotDiffRef:
                line = plt.plot(q,diffs_abs[idiff],color=color)[0]
                lines_abs.append(line)
    if title is not None: fig.axes[0].set_title(title)
    legend = plt.legend(loc=4)
    plt.grid()
    plt.xlabel(r"q ($\AA^{-1}$)")
    # we will set up a dict mapping legend line to orig line, and enable
    # picking on the legend line
    lined = dict()
    for legline, origline in zip(legend.get_lines(), lines_diff):
        legline.set_picker(5)  # 5 pts tolerance
        lined[legline] = origline
  
    def onpick(event):
        # on the pick event, find the orig line corresponding to the
        # legend proxy line, and toggle the visibility
        legline = event.artist
        origline = lined[legline]
        vis = not origline.get_visible()
        origline.set_visible(vis)
        # Change the alpha on the line in the legend so we can see what lines
        # have been toggled
        if vis:
            legline.set_alpha(1.0)
        else:
            legline.set_alpha(0.2)
        fig    = plt.gcf() 
        fig.canvas.draw()
 
    fig.canvas.mpl_connect('pick_event', onpick)
    return lines_diff,lines_abs

def updateLines(lines,data):
    for l,d in zip(lines,data):
        l.set_ydata(d)



def saveTxt(fname,q,data,headerv=None,info=None,overwrite=True,columns=''):
  """ Write data to file 'fname' in text format.
      Inputs:
        q = x vector
        data = one or 2D array (first axis is q)
        info = dictionary (saved as '# key : value') or string
        headerv = vector to be used as header or string
  """
  if os.path.isfile(fname) and not overwrite:
    log.warn("File %s exists, returning",fname)
    return
  if isinstance(info,dict):
    header = [ "# %s : %s" %(k,str(v)) for (k,v) in info.items() ]
    header = "\n".join(header); # skip first #, will be added by np
  elif isinstance(info,str):
    header = info
  else:
    header = ""
  if isinstance(headerv,str): header += "\n%s" % headerv
  if data.ndim == 1:
    x = np.vstack( (q,data) )
  elif data.ndim == 2:
    x = np.vstack( (q,data) )
    if headerv is not None:
      headerv = np.concatenate(( (data.shape[1],),headerv))
      x = np.hstack( (headerv[:,np.newaxis],x) )
  if columns != '':
    s = "#" + " ".join( [str(c).center(12) for c in columns] )
    header = header + "\n" + s if header != '' else s
  np.savetxt(fname,x.T,fmt="%+10.5e",header=header,comments='')

def reshapeToBroadcast(what,ref):
  """ expand the 1d array 'what' to allow broadbasting to match 
      multidimentional array 'ref'. The two arrays have to same the same 
      dimensions along the first axis
  """
  if what.shape == ref.shape: return what
  assert what.shape[0] == ref.shape[0],\
    "automatic reshaping requires same first dimention"
  shape  = [ref.shape[0],] + [1,]*(ref.ndim-1)
  return what.reshape(shape)

def radToQ(theta,**kw):
  """ theta is the scattering angle (theta_out - theta_in);
      kw should be have either E or wavelength
      it returns the scattering vector in the units of wavelength """
  # Energy or wavelength should be in kw
  assert "E" in kw or "wavelength" in kw,\
    "need wavelength or E to convert rad to Q"
  # but not both
  assert not ("E" in kw and "wavelength" in kw),\
    "conflicting arguments (E and wavelength)"
  if "E" in kw: kw["wavelength"] = 12.398/kw["E"]
  return 4*np.pi/kw["wavelength"]*np.sin(theta/2)

def degToQ(theta,**kw):
  theta = theta/180.*np.pi
  return radToQ(theta,**kw)
degToQ.__doc__ = radToQ.__doc__

def qToTwoTheta(q,asDeg=False,**kw):
  """ Return scattering angle from q (given E or wavelength) """
  # Energy or wavelength should be in kw
  assert "E" in kw or "wavelength" in kw,\
    "need wavelength or E to convert rad to Q"
  # but not both
  assert not ("E" in kw and "wavelength" in kw),\
    "conflicting arguments (E and wavelength)"
  if "E" in kw: kw["wavelength"] = 12.398/kw["E"]
  theta = 2*np.arcsin(q*kw["wavelength"]/4/np.pi)
  if asDeg: theta = np.rad2deg(theta)
  return theta

def attenuation_length(compound, density=None, natural_density=None,energy=None, wavelength=None):
  """ extend periodictable.xsf capabilities """
  import periodictable.xsf
  if energy is not None: wavelength = periodictable.xsf.xray_wavelength(energy)
  assert wavelength is not None, "scattering calculation needs energy or wavelength"
  if (np.isscalar(wavelength)): wavelength=np.array( [wavelength] )
  n = periodictable.xsf.index_of_refraction(compound=compound,
                           density=density, natural_density=natural_density,
                           wavelength=wavelength)
  attenuation_length = (wavelength*1e-10)/ (4*np.pi*np.imag(n))
  return np.abs(attenuation_length)

def transmission(material='Si',thickness=100e-6, density=None, natural_density=None,energy=None, wavelength=None):
  """ extend periodictable.xsf capabilities """
  att_len = attenuation_length(material,density=density,
            natural_density=natural_density,energy=energy,wavelength=wavelength)
  return np.exp(-thickness/att_len)


def _phosphorAbsorption(twotheta,mu=1770,thickness=40e-6,E=None):
  """ return 1-np.exp(-mu*thick/np.cos(np.deg2rad(theta)))
      - mu is the neperian absortption linear coefficient (m-1)
      - twotheta is the scattering angle (in degrees)
  """
  if mu =='auto':
    assert E is not None, "phosphorAbsorption with automatic mu requires E"
    mu = 1/attenuation_length("Ce",energy=E,density=0.475)
  return 1-np.exp(-mu*thickness/np.cos(np.deg2rad(twotheta)))

def phosphorCorrection(twotheta,mu=1770,thickness=40e-6,E=None,normalizeToZeroAngle=False):
  """ helper function to correct for angle dependent absorption of the phosphor screen for an
      x-ray detector.
      return the correction factor one has to *multiply* the data for.
      - mu is the neperian absortption linear coefficient (m-1) [could be 'auto' if E is given]
      - twotheta is the scattering angle (in degrees)
  """
  corr = 1/_phosphorAbsorption(twotheta,mu=mu,thickness=thickness,E=E)
  if normalizeToZeroAngle: corr = corr*_phosphorAbsorption(0,mu=mu,thickness=thickness,E=E)
  return corr




def chargeToPhoton(chargeOrCurrent,material="Si",thickness=100e-6,energy=10,e_hole_pair=3.63):
  """
    Function to convert charge (or current to number of photons (or number 
    of photons per second)
    
    Parameters
    ----------
    
    chargeOrCurrent: float or array
    material : str
       Used to calculate 
     
  """
  # calculate absortption
  A = 1-transmission(material=material,energy=energy,thickness=thickness)
  chargeOrCurrent = chargeOrCurrent/A

  e_hole_pair_energy_keV = e_hole_pair*1e-3
  n_charge_per_photon = energy/e_hole_pair_energy_keV
  # convert to Q
  charge_per_photon = n_charge_per_photon*1.60217662e-19
  nphoton = chargeOrCurrent/charge_per_photon
 
  if len(nphoton) == 1: nphoton = float(nphoton)
  return nphoton

def approx(values,approx_values):
    """ returns array where every value is replaced by the closest in approx_values

        This funciton is useful for rebinning

        Example:
        -------
        approx( np.arange(0,1,0.1), [0,0.3,0.7] )
           array([ 0. ,  0. ,  0.3,  0.3,  0.3,  0.7,  0.7,  0.7,  0.7,  0.7])

    """
    # make sure they are arrays
    values = np.asarray(values)
    approx_values = np.asarray(approx_values)
    # create outter difference
    diff = np.abs(values[:,np.newaxis] - approx_values)
    args = np.argmin(diff,axis=1)
    values = approx_values[args]
    #values = np.asarray( [ approx_values[np.argmin(np.abs(v-approx_values))] for v in values] )
    return values
  
def logToScreen():
  """ It allows printing to terminal on top of logfile """
  # define a Handler which writes INFO messages or higher to the sys.stderr
  console = logging.StreamHandler()
  console.setLevel(logging.INFO)
  # set a format which is simpler for console use
  formatter = logging.Formatter('%(message)s')
  # tell the handler to use this format
  console.setFormatter(formatter)
  # add the handler to the root logger (if needed)
  if len(logging.getLogger('').handlers)==1:
    logging.getLogger('').addHandler(console)





from setuptools import setup, find_packages

def readme():
    with open('README.rst') as f:
        return f.read()

setup(name='trx',
      version='0.5.12',
      description='tools for (T)ime (R)esolved (X)-ray techniques',
      long_description=readme(),
      url='https://github.com/marcocamma/trx',
      author='marco cammarata',
      author_email='marcocamma@gmail.com',
      license='MIT',
#      packages=['trx'],
      packages=find_packages("."),
      install_requires=[
          'numpy',
          'fabio',
          'pyFAI',
          'matplotlib',
          'datastorage>=0.4.3',
          'statsmodels'
      ],
      zip_safe=False)



# -*- coding: utf-8 -*-
from __future__ import print_function,division,absolute_import

import logging
log = logging.getLogger(__name__)  # __name__ is "foo.bar" here

import numpy as np
np.seterr(all='ignore')
from .string import timeToStr

try:
  import matplotlib.pyplot as plt
except ImportError:
  log.warn("Can't import matplotlib !")

def plotdata(data,x=None,plotAverage=True,showTrend=True,title=None,\
      clim='auto',fig=None):
    """ Plot 2D array to look for trends/stability or outlier

        It usually works best with normalized data

        Parameters
        ----------
        data: dict-like object or 2D array.
          if dict-like it has to have keys
          'q' used for one axis
          'data' for 2D array[numimg,nQ]
        x: array or None
          if given it has to have a length of numimg
          if None will be 0,1,2,3,...numimg and interpreted
          as image number
        plotAverage: bool
          if True, plot average
        showTrend: bool
          if True, show 2D plot (most common use)
        clim: 'auto' or tuple
          color scale to use, is 'auto' 1.5%,98.5% percentiles are used
        fig:
          matplotlib fig instance to use, if None creates a new one
        title: str
          title to use for plot (if None and data['folder'] exists
          the value is used as title
          
    """
    if isinstance(data,np.ndarray):
      q = np.arange( data.shape[-1] )
    else:
      if title is None and 'folder' in data: title = data["folder"]
      q = data["q"]
      data = data["data"]
    if not (plotAverage or showTrend): return

    if x is None: x = np.arange(data.shape[0])
    if clim == 'auto': clim = np.nanpercentile(data,(1.5,98.5))
    one_plot = showTrend or plotAverage
    two_plot = showTrend and plotAverage

    if one_plot and not two_plot:
      if fig is None:
        fig,ax = plt.subplots(1,1)
      else:
        fig.clear()
        ax = fig.axes
    if two_plot:
      if fig is None:
        fig,ax = plt.subplots(2,1,sharex=True)
      else:
        ax = fig.axes
 
    ax = np.atleast_1d(ax)
    if showTrend:
      plt.sca(ax[1])
      plt.pcolormesh(q,x,data)
      plt.ylabel("image number, 0 being older")
      plt.xlabel(r"q ($\AA^{-1}$)")
      plt.clim( *clim )
    if plotAverage:
        ax[0].plot(q,np.nanmean(data,axis=0))
    if (plotAverage or showTrend) and title is not None:
      plt.title(title)

def plotdiffs(data,select=None,err=None,absSignal=None,absSignalScale=10,
           showErr=False,cmap=plt.cm.jet,fig=None,title=None,plotDiffRef=False):
    """ Plot difference data

        Parameters
        ----------
        data: dict-like object or (q,diffs,scan)tuple
          if dict-like it has to have keys
          'q' used for one axis
          'diffs' for 2D array[numscan,nQ]
          'scan'  scan points
          if present 'err' is used for errorbars
          if present 'diffs_plus_ref' is used
        select: slice
          to select time delays (for example select=slice(None,None,2))
        absSignal: array[nQ]
          if present, ass this data to the plot (could be the average)
        absSignalScale: float
          the factor to use to divide the absSignal for, it is kept as
          different parameters because the scaling it is added to the legend
        showErr: bool
          if True, plot errorbars as well
        cmap: matplotlib color map
          color map to use
        fig:
          matplotlib fig instance to use, if None creates a new one
        title: str
          title to use for plot (if None, no title is added)
        plotDiffRef:
          plot not only the diff but also the diffs_plus_ref (if present in data)
    """
    if isinstance(data,(list,tuple)):
        q,diffs,scan = args
        diffs_abs = None
    else:
        q = data["q"]
        diffs = data["diffs"]
        scan  = data["scan"]
        err   = data.get("err",None)
        diffs_abs = data.get("diffs_plus_ref",None)

    # this selection trick done in this way allows to keep the same colors when 
    # subselecting (because I do not change the size of diffs)
    if select is not None:
        indices = range(*select.indices(scan.shape[0]))
    else:
        indices = range(len(scan))
 
    if fig is None: fig = plt.figure()
 
    lines_diff = []
    lines_abs = []
    if absSignal is not None:
        line = plt.plot(q,absSignal/absSignalScale,lw=3,
                      color='k',label="absSignal/%s"%str(absSignalScale))[0]
        lines_abs.append(line)
    for linenum,idiff in enumerate(indices):
        color = cmap(idiff/(len(diffs)-1))
        label = timeToStr(scan[idiff])
        kw = dict( color = color, label = label )
        if err is not None and showErr:
            line = plt.errorbar(q,diffs[idiff],err[idiff],**kw)[0]
            lines_diff.append(line)
        else:
            line = plt.plot(q,diffs[idiff],**kw)[0]
            lines_diff.append(line)
            if diffs_abs is not None and plotDiffRef:
                line = plt.plot(q,diffs_abs[idiff],color=color)[0]
                lines_abs.append(line)
    if title is not None: fig.axes[0].set_title(title)
    legend = plt.legend(loc=4)
    plt.grid()
    plt.xlabel(r"q ($\AA^{-1}$)")
    # we will set up a dict mapping legend line to orig line, and enable
    # picking on the legend line
    lined = dict()
    for legline, origline in zip(legend.get_lines(), lines_diff):
        legline.set_picker(5)  # 5 pts tolerance
        lined[legline] = origline
  
    def onpick(event):
        # on the pick event, find the orig line corresponding to the
        # legend proxy line, and toggle the visibility
        legline = event.artist
        origline = lined[legline]
        vis = not origline.get_visible()
        origline.set_visible(vis)
        # Change the alpha on the line in the legend so we can see what lines
        # have been toggled
        if vis:
            legline.set_alpha(1.0)
        else:
            legline.set_alpha(0.2)
        fig    = plt.gcf() 
        fig.canvas.draw()
 
    fig.canvas.mpl_connect('pick_event', onpick)
    return lines_diff,lines_abs

def updateLines(lines,data):
    for l,d in zip(lines,data):
        l.set_ydata(d)



# -*- coding: utf-8 -*-
""" 
    Module to handle masks: it defines:

      - Mask: a mask object with command line methods: addPolygon, etc
      - makeMaskGui: a GUI based way of creating masks
      - maskBorder: to mask the borders of an array
      - maskCenterLines: to mask the central lines (good for multi-panel det)
      - interpretMask: interpret mask element (filename,y>500,array)
      - interpretMasks: add up list of mask elements
"""
from __future__ import print_function,division,absolute_import
import sys
if sys.version_info.major == 2: input=raw_input
import logging
log = logging.getLogger(__name__)

import os
import re
import collections
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.path import Path
import fabio

def read(fname):
  """ read data from file using fabio """
  f = fabio.open(fname)
  data = f.data
  del f; # close file
  return data


maskComponent = collections.namedtuple('maskComponent',['operation','geometry','vertices'])

def _rectangleToMask(X,Y,vertices):
    ( (x1,y1), (x2,y2) ) = vertices
    if x1>x2: x1,x2=x2,x1
    if y1>y2: y1,y2=y2,y1
    return (X>x1) & (X<x2) & ( Y>y1) & (Y<y2)

def _circleToMask(X,Y,vertices):
    c,p = vertices
    r = np.sqrt( (p[0]-c[0])**2 + (p[1]-c[1])**2 )
    d = np.sqrt((X-c[0])**2+(Y-c[1])**2)
    return d<r

def _polygonToMask(X,Y,vertices):
    points = np.vstack((X.flatten(),Y.flatten())).T
    path = Path(vertices)
    grid = path.contains_points(points)
    return grid.reshape(X.shape)

class Mask(object):
  """ class for making masks. True are pixels masked OUT.

      This class provides methods for adding/subtracting components in the
      for of rectangles, circles, polygons. Each addition/subtraction is 
      stored as 'operations' that are then applied on request.
      
      Parameters
      ----------
      img: {array,filename,shape of array}
          img to use for calculation. Since only the shape is used a shape-like
          tuple can be provided


  """
  def __init__(self,img=None):
    self.comp = []
    if img is not None:
      self.shape = img.shape
    else:
      self.shape = None
    if isinstance(img,str): img = read(img)
    # if img is not array at this point assume it is a shape-tuple
    if not isinstance(img,np.ndarray): img = np.zeros( img, dtype=bool )
    self.img  = img
    self.mask = None
    self._cache = None

  def _define_component(self,operation,geometry,*vertices):
    #print("define_comp",type(vertices),vertices)
    if geometry == 'circle' and len(vertices) == 3:
      xcen,ycen,radius = vertices
      vertices = ( (xcen,ycen), (xcen+radius,ycen) )
    if geometry == 'rectangle' and len(vertices) == 4:
      vertices = ( (vertices[0],vertices[1]),(vertices[2],vertices[3]) )
    # make sure vertices tuples
    if isinstance(vertices,list):
      vertices = [ (v[0],v[1]) for v in vertices ]
      vertices = tuple(vertices)
    a = dict( vertices = None )
    self.comp.append( maskComponent(operation=operation,vertices=vertices,geometry=geometry) )
    
  
  def addCircle(self,*vertices): self._define_component( 'add', 'circle', *vertices )
  def subtractCircle(self,*vertices): self._define_component( 'subtract', 'circle', *vertices )

  def addRectangle(self,*vertices): self._define_component( 'add','rectangle', *vertices)
  def subtractRectangle(self,*vertices): self._define_component( 'subtract','rectangle',*vertices)
  
 
  def addPolygon(self,*vertices): self._define_component( 'add','polygon',*vertices)
  def subtractPolygon(self,*vertices): self._define_component( 'subtract','polygon',*vertices)


  def getMask(self,shape=None):
    if shape is None and self.img is not None: shape = self.img.shape
    if shape is None and self.img is     None: shape = self._cache['shape']
    
    if self._cache is None: self._cache = dict( shape = shape )
    # reset cache if shape does not match
    if shape != self._cache['shape']:
      self._cache = dict( shape = shape )
    X,Y = np.meshgrid ( range(shape[1]),range(shape[0]) )
    for component in self.comp:
      if component not in self._cache:
        if component.geometry == 'circle':
          mask = _circleToMask( X,Y,component.vertices )
        elif component.geometry == 'rectangle':
          mask = _rectangleToMask( X,Y,component.vertices )
        elif component.geometry == 'polygon':
          mask = _polygonToMask( X,Y,component.vertices )
        else:
          raise ValueError("Mask type %s not recongnized"%component.geometry)
        self._cache[component] = mask
    mask = np.zeros(shape,dtype=np.bool)
    for comp in self.comp:
      m = self._cache[ comp ]
      if (comp.operation == "add"):
        mask[m] = True
      else:
        mask[m] = False
    self.mask = mask
    return mask

  def getMatplotlibMask(self,shape=None):
    mask = self.getMask(shape=shape)
    # convert
    mpl_mask = np.zeros( (mask.shape[0],mask.shape[1],4) )
    mpl_mask[:,:,:3] = 0.5; # gray color
    mpl_mask[:,:,3] = mask/2; # give some transparency
    return mpl_mask

  def save(self,fname,inverted=False):
    import fabio
    if self.mask is None: self.getMask()
    mask = self.mask
    if (inverted): mask = ~mask
    if os.path.splitext(fname)[1] == ".npy":
      np.save(fname,mask)
    else:
      i=fabio.edfimage.edfimage(mask.astype(np.uint8)); # edf does not support bool
      i.save(fname)

def snap(point,shape,snapRange=20):
  """ snap 'point' if within 'snapRange' from the border defined by 'shape' """
  snapped = list(point)
  if snapped[0] < snapRange: snapped[0] = 0
  if snapped[0] > shape[1]-snapRange: snapped[0] = shape[1]
  if snapped[1] < snapRange: snapped[1] = 0
  if snapped[1] > shape[0]-snapRange: snapped[1] = shape[0]
  return tuple(snapped)

def getPoints(N=1,shape=(100,100),snapRange=0):
  if N<1: print('Right click cancels last point, middle click ends the polygon')
  c = plt.ginput(N)
  c = [ snap(point,shape,snapRange=snapRange) for point in c ]
  if len(c) == 1: c = c[0]
  return c

def makeMaskGui(img,snapRange=60,clim='auto'):
  """ interactive, click based approach do define a mask.
      
      Parameters
      ----------
      snapRange : int
          controls border snapping (in pixels) use <= 0 to disable;
      clim: 'auto' or list(min,max)
          controls color scale os image, if 'auto' uses 2%-98% percentile

      Returns
      -------
      Mask
          instance of the Mask class that allows to modify or save the mask

  """
  if isinstance(img,str): img = read(img)
  mask = Mask(img)
  if clim == "auto": clim = np.percentile(img,(2,98))
  ans='ok'
  while (ans != 'done'):
    plt.imshow(img)
    plt.clim(clim)
    plt.imshow(mask.getMatplotlibMask())
    plt.pause(0.01)
    ans = input("What's next p/P/c/C/r/R/done? (capitals = subtract)")
    if ans == "c":
      print("Adding circle, click on center then another point to define radius")
      vertices = getPoints(N=2,shape=img.shape,snapRange=snapRange)
      mask.addCircle(*vertices)
    if ans == "C":
      print("Subtracting circle, click on center then another point to define radius")
      vertices = getPoints(N=2,shape=img.shape,snapRange=snapRange)
      mask.subtractCircle(*vertices)
    if ans == "r":
      print("Adding rectangle, click on one corner and then on the opposite one")
      vertices = getPoints(N=2,shape=img.shape,snapRange=snapRange)
      mask.addRectangle(*vertices)
    if ans == "R":
      print("Subtracting rectangle, click on one corner and then on the opposite one")
      vertices = getPoints(N=2,shape=img.shape,snapRange=snapRange)
      mask.subtractRectangle(*vertices)
    if ans == 'p':
      print("Adding polygon")
      vertices = getPoints(N=-1,shape=img.shape,snapRange=snapRange)
      mask.addPolygon(*vertices)
    if ans == 'P':
      print("Subtracting polygon")
      vertices = getPoints(N=-1,shape=img.shape,snapRange=snapRange)
      mask.subtractPolygon(*vertices)

    plt.imshow(mask.getMatplotlibMask())
    plt.pause(0.01)
  fname = input("Enter a valid filename (ext .edf or .npy) if you want to save the mask (empty otherwise)")
  try:
    if fname != '':
      ext = os.path.splitext(fname)[1]
      if ext == '.edf':
        mask.save(fname)
      elif ext == '.npy':
        np.save(fname,mask.getMask())
  except Exception as e:
    log.error("Error in saving mask")
    log.error(e)
  finally:
    return mask
    
def maskBorder(width,shape):
  """ mask the border of an array for a given width
      
      Parameters
      ----------
      width : int
          the width of the region to mask (>0)

      Returns
      -------
      boolean (False/True) array
          True are the pixels masked out
  """
  assert isinstance(width,int), "width has to be integer"
  assert width>0, "width has to be positive"
  mask = np.zeros(shape,dtype=bool)
  mask[ :width  ,   :     ] = True
  mask[ -width: ,   :     ] = True
  mask[    :    , :width  ] = True
  mask[    :    , -width: ] = True
  return mask

def maskCenterLines(width,shape):
  """ mask a cross going trough the center of the array for a given width
      
      Parameters
      ----------
      width : int
          the width of the region to mask (>0)

      Returns
      -------
      boolean (False/True) array
          True are the pixels masked out
  """
  assert isinstance(width,int), "width has to be integer"
  assert width>0, "width has to be positive"
  mask = np.zeros(shape,dtype=bool)
  if isinstance(width,int): width = (width,width)
  c0 = int(shape[0]/2)
  c1 = int(shape[1]/2)
  w0 = int(width[0]/2)
  w1 = int(width[1]/2)
  mask[ c0-w0:c0+w0  ,       :      ] = True
  mask[      :       ,  c1-w1:c1+w1 ] = True
  return mask

g_mask_str = re.compile("(\w)\s*(<|>)\s*(\d+)")

def interpretMask(mask,shape=None):
  """ Interpret 'mask' as a mask

      Parameters
      ----------
      mask : filename or array or string like y>500

      shape : array or tuple
          needed to interpret y>500

      Returns
      -------
      boolean (False/True) array
          True are the pixels masked out
  """
  maskout = None
  ## simplest case, an existing file
  if isinstance(mask,str) and os.path.isfile(mask):
    maskout = read(mask).astype(np.bool)
  ## mask string
  elif isinstance(mask,str) and not os.path.isfile(mask):
    if isinstance(shape,np.ndarray) : shape = shape.shape
    err_msg = ValueError("The string '%s' could not be interpreted as simple\
              mask; it should be something like x>10"%mask)
    assert shape is not None, "_interpretMask needs a shape to interpret a string"
    # interpret string
    maskout = np.zeros(shape,dtype=bool)
    match = g_mask_str.match(mask)
    if match is None: raise err_msg
    (axis,sign,lim) = match.groups()
    if axis not in ("x","y"): raise err_msg
    if sign not in (">","<"): raise err_msg
    lim = int(lim)
    idx = slice(lim,None) if sign == ">" else slice(None,lim)
    if axis == 'y':
      maskout[idx,:] = True
    else:
      maskout[:,idx] = True
  elif isinstance(mask,np.ndarray):
    maskout = mask.astype(np.bool)
  elif mask is None:
    assert shape is not None, "_interpretMask needs a shape to interpret a string"
    maskout = np.zeros(shape,dtype=bool)
  else:
    maskout = None
    raise ValueError("Could not interpret %s as mask input"%mask)
  
  if shape is not None and maskout.shape != shape:
    raise ValueError("The mask shape %s does not match the shape given as\
      argument %s"%(maskout.shape,shape))
  return maskout

def interpretMasks(masks,shape=None):
  """ Interpret a single or a list of mask elements

      Every element can be an array, a filename to read, a 'mask string'
      (y>500). 

      Parameters
      ----------
      masks : a 'mask element' or a list of mask elements

      shape : array or tuple
          needed to interpret y>500

      Returns
      -------
      boolean (False/True) array
          True are the pixels masked out
 
  """
  if isinstance(masks,np.ndarray): return masks.astype(bool)
  # make iterable
  if not isinstance( masks, (list,tuple,np.ndarray) ): masks = (masks,)
  masks = [interpretMask(mask,shape) for mask in masks]
  # put them all together
  mask = masks[0]
  for m in masks[1:]:
    mask = np.logical_or(mask,m)
  return mask


def test(shape=(1000,2000)):
  """ Make a simple mask programmatically """
  mask = Mask()
  mask.addCircle(400,300,250)
  mask.subtractCircle(400,300,150)
  mask.addRectangle(350,250,1500,700)
  plt.imshow( mask.getMask(shape) )
  return mask



if __name__ == "__main__":
  test()
  plt.show()
  ans=input("Enter to finish") 



# -*- coding: utf-8 -*-
from __future__ import print_function,division,absolute_import

import logging
log = logging.getLogger(__name__)

try:
  import lmfit
except ImportError as e:
  log.error("To use this submodule, please install `lmfit`")
import numpy as np
import logging as log

def fitPeak(x,y,err=1,autorange=False):
  pv = lmfit.models.PseudoVoigtModel()
  if isinstance(err,np.ndarray):
    if np.all(err==0):
      err = 1
      log.warn("Asked to fit peak but all errors are zero, forcing them to 1")
    elif np.isfinite(err).sum() != len(err):
      idx = np.isfinite(err)
      x = x[idx]
      y = y[idx]
      err = err[idx]
      log.warn("Asked to fit peak but some errorbars are infinite or nans,\
                excluding those points")
  if autorange:
    # find fwhm
    idx = np.ravel(np.argwhere( y<y.max()/2 ))
    # find first crossing
    p1 = idx[idx<np.argmax(y)][-1]
    p2 = idx[idx>np.argmax(y)][0]
    c = int( (p1+p2)/2 )
    dp = int( np.abs(p1-p2) )
    idx = slice(c-dp,c+dp)
    x = x[idx]
    y = y[idx]
  pars = pv.guess(y,x=x)
  ret  = pv.fit(y,x=x,weights=1/err,params=pars)
  return ret



# -*- coding: utf-8 -*-
from __future__ import print_function,division,absolute_import

import logging
log = logging.getLogger(__name__)

import numpy as np
np.seterr(all='ignore')
from . import utils
from . import filters
from datastorage import DataStorage
import os

def subtractReferences(i,idx_ref, useRatio = False):
  """ given data in i (first index is shot num) and the indeces of the 
      references (idx_ref, array of integers) it interpolates the closest
      reference data for each shot and subtracts it (or divides it, depending
      on useRatio = [True|False]; 
      Note: it works in place (i.e. it modifies i) """
  iref=np.empty_like(i)
  idx_ref = np.squeeze(idx_ref)
  idx_ref = np.atleast_1d(idx_ref)
  # sometime there is just one reference (e.g. sample scans)
  if idx_ref.shape[0] == 1:
    if useRatio:
      return i/i[idx_ref]
    else:
      return i-i[idx_ref]
  # references before first ref are "first ref"
  iref[:idx_ref[0]] = i[idx_ref[0]]
  # references after last ref are "last ref"
  iref[idx_ref[-1]:] = i[idx_ref[-1]]
  _ref = 0
  for _i in range(idx_ref[0],idx_ref[-1]):
    if _i in idx_ref: continue
    idx_ref_before = idx_ref[_ref]
    idx_ref_after  = idx_ref[_ref+1]
    ref_before = i[idx_ref_before]
    ref_after  = i[idx_ref_after]
    weight_before = float(_i-idx_ref_before)/(idx_ref_after-idx_ref_before)
    weight_after  = 1-weight_before
    # normal reference for an on chi, the weighted average
    iref[_i]      = weight_before*ref_before + weight_after*ref_after
    if _i>=idx_ref_after-1: _ref += 1
    log.debug("For image %d : %d-%d"%(_i,idx_ref_before,idx_ref_after))
  # take care of the reference for the references ...
  if len(idx_ref) >  2:
    iref[idx_ref[0]] = i[idx_ref[1]]
    iref[idx_ref[-1]] = i[idx_ref[-2]]
    for _i in range(1,len(idx_ref)-1):
      idx_ref_before = idx_ref[_i-1]
      idx_ref_after  = idx_ref[_i+1]
      ref_before = i[idx_ref_before]
      ref_after  = i[idx_ref_after]
      weight_before = float(idx_ref[_i]-idx_ref_before)/(idx_ref_after-idx_ref_before)
      weight_after  = 1-weight_before
      # normal reference for an on chi, the weighted average
      iref[idx_ref[_i]]    = weight_before*ref_before + weight_after*ref_after
      log.debug("For reference image %d : %d-%d"%(idx_ref[_i],idx_ref_before,idx_ref_after))
  else:
    #print(idx_ref)
    #print(iref[idx_ref])
    iref[idx_ref]=i[idx_ref[0]]
    #print(iref[idx_ref])
  if useRatio:
    i /= iref
  else:
    i -= iref
  return i

def averageScanPoints(scan,data,errAbs=None,isRef=None,lpower=None,
    useRatio=False,funcForAveraging=np.nanmean,chi2_0_max='auto'):
  """ Average data for equivalent values in 'scan' array

      given scanpoints in 'scan' and corresponding data in 'data'
      average all data corresponding the exactly the same scanpoint.
      If the values in scan are coming from a readback, rounding might be
      necessary.

      Parameters
      ----------
      scan : array(N)
          array of scan points
      data : array(N,M)
          array of data to average, first axis correspond to scan index
      errAbs : None or array as data
          errbar for each data point. if None take the standard deviation 
          over images in given scan point
      isRef : None or array(N)
          if None no reference is subtracted. if array, True indicate that 
          a particular image is a reference one
      lpower : None or array(N)
          if not None, time resolved difference or ratio is normalized by it
      useRatio : bool
          use True if you want to calculate ratio ( I_{on}/I_{ref} ) instead
          of I_{on} - I_{off}
      funcForAveraging: function accepting axis=int keyword argument
          is usually np.nanmean or np.nanmedian.
      chi2_0_max = None, "auto" or float
          simple chi2_0 threshold filter. use trx.filters for more advanced
          ones. If auto, define max as 95% percentle. if None it is not applied

      Returns
      -------
      DataStorage instance with all info
"""
  args = dict( isRef = isRef, lpower = lpower, useRatio = useRatio )
  data = data.astype(np.float)
  average = np.mean(data,axis=0)
  median  = np.median(data,axis=0)

  if isRef is None: isRef = np.zeros( data.shape[0], dtype=bool )
  isRef = np.asarray(isRef).astype(bool)
  assert data.shape[0] == isRef.shape[0], \
    "Size mismatch, data is %d, isRef %d"%(data.shape[0],isRef.shape[0])

  # subtract reference only is there is at least one
  if isRef.sum()>0:
    # create a copy (subtractReferences works in place)
    diff_all = subtractReferences(data.copy(),np.argwhere(isRef),
               useRatio=useRatio)
    ref_average = funcForAveraging(data[isRef],axis=0)
  else:
    diff_all = data
    ref_average = np.zeros_like(average)

  # normalize signal for laser intensity if provided
  if lpower is not None:
    lpower = utils.reshapeToBroadcast(lpower,data)
    if useRatio is False:
      diff_all /= lpower
    else:
      diff_all = (diff_all-1)/lpower+1

  scan_pos = np.unique(scan)
  shape_out = [len(scan_pos),] + list(diff_all.shape[1:])
  diffs     = np.empty(shape_out)
  diff_err  = np.empty(shape_out)
  diffs_in_scan = []
  chi2_0 = []
  for i,t in enumerate(scan_pos):
    shot_idx = (scan == t) & ~isRef
    if shot_idx.sum() == 0:
      log.warn("No data to average for scan point %s"%str(t))

    # select data for the scan point
    diff_for_scan = diff_all[shot_idx]
    if errAbs is not None:
      noise  = np.nanmean(errAbs[shot_idx],axis = 0)
    else:
      noise = np.nanstd(diff_for_scan, axis = 0)

    # if it is the reference take only every second ...
    if np.all( shot_idx == isRef ):
      diff_for_scan = diff_for_scan[::2]

    diffs_in_scan.append( diff_for_scan )

    # calculate average
    diffs[i] = funcForAveraging(diff_for_scan,axis=0)

    # calculate chi2 of different repetitions
    chi2 = np.power( (diff_for_scan - diffs[i])/noise,2)
    # sum over all axis but first
    for _ in range(diff_for_scan.ndim-1):
      chi2 = np.nansum( chi2, axis=-1 )

    # store chi2_0
    chi2_0.append( chi2/diffs[i].size )

    # store error of mean
    diff_err[i] = noise/np.sqrt(shot_idx.sum())
  ret = dict(scan=scan_pos,diffs=diffs,err=diff_err,
        chi2_0=chi2_0,diffs_in_scan=diffs_in_scan,
        ref_average = ref_average, diffs_plus_ref=diffs+ref_average,
        average=average,median=median,args=args)
  ret = DataStorage(ret)
  if chi2_0_max is not None:
    ret = filters.chi2Filter(ret,threshold=chi2_0_max)
    ret = filters.applyFilters(ret)
  return ret


def calcTimeResolvedSignal(scan,data,err=None,reference="min",q=None,
    monitor=None,saveTxt=True,folder=os.curdir,**kw):
  """
    reference: can be 'min', 'max', a float|integer or an array of booleans
    data     : >= 2dim array (first index is image number)
    q        : is needed if monitor is a tuple|list
    saveTxt  : will save txt outputfiles (diff_av_*) in folder
    other keywords are passed to averageScanPoints

    **kw are passed to averageScanPoints. Notable parameters are
      isRef : None or array(N); if None no reference is subtracted
      lpower : None or array(N)
      useRatio : bool
      chi2_0_max = None, "auto" or float
  """
  if isinstance(reference,str) and reference == "min":
    isRef = (scan == scan.min())
  elif isinstance(reference,str) and reference == "max":
    isRef = (scan == scan.max())
  elif isinstance(reference,(float,int)):
    isRef = (scan == reference)
  else:
    isRef = reference
  # normalize if needed
  if monitor is not None:
    if isinstance(monitor,(tuple,list)):
      assert q is not None, "q is None and can't work with q-range scaling"
      assert data.ndim == 2, "currently q-range scaling works with 2dim data"
      idx = (q>= monitor[0]) & (q<= monitor[1])
      monitor = np.nanmedian(data[:,idx],axis=1)
    monitor = utils.reshapeToBroadcast(monitor,data)
    data = data/monitor
    if err is not None: err  = err/monitor
  ret = averageScanPoints(scan,data,errAbs=err,isRef=isRef,**kw)
  if q is not None: ret["q"] = q
  return ret

def saveTxt(folder,data,delayToStr=True,basename='auto',info="",**kw):
  """ data must be a DataStorage instance """
  # folder ends usually with sample/run so use the last two subfolders
  # the abspath is needed in case we analyze the "./"
  folder = os.path.abspath(folder);
  if basename == 'auto':
      sep = os.path.sep
      basename = "_".join(folder.rstrip(sep).split(sep)[-2:]) + "_"
  q = data.q if "q" in data else np.arange(data.diffs.shape[-1])
  # save one file with all average diffs
  fname = os.path.join(folder,"%sdiff_av_matrix.txt" %basename)
  utils.saveTxt(fname,q,data.diffs,headerv=data.scan,**kw)
  fname = os.path.join(folder,"%sdiff_plus_ref_av_matrix.txt" %basename)
  utils.saveTxt(fname,q,data.diffs_plus_ref,headerv=data.scan,**kw)
  # save error bars in the matrix form
  fname = os.path.join(folder,"%sdiff_av_matrix_err.txt" % basename)
  utils.saveTxt(fname,q,data.err,headerv=data.scan,**kw)

  for iscan,scan in enumerate(data.scan):
    scan = utils.timeToStr(scan) if delayToStr else "%+10.5e" % scan

    # try retreiving info on chi2
    try:
      chi2_0  = data.chi2_0[iscan]
      info_delay = [ "# rep_num : chi2_0 , discarded by chi2masking ?", ]
      for irep,value in enumerate(chi2_0):
        info_delay.append( "# %d : %.3f" % (irep,value))
        if 'chi2' in data.masks: info_delay[-1] += " %s"%str(data.masks['chi2'][iscan][irep])
      info_delay = "\n".join(info_delay)
      if info != '': info_delay = "%s\n%s" % (info,info_delay)
    except AttributeError:
      info_delay = info

    # save one file per timedelay with average diff (and err)
    fname = os.path.join(folder,"%sdiff_av_%s.txt" %(basename,scan))
#    if 'mask' in data:
#      tosave = np.vstack( (data.diffs[iscan],data.err[iscan],
#               data.dataUnmasked[iscan],data.errUnmasked[iscan] ) )
#      columns = 'q diffmask errmask diffnomask errnomask'.split()
#    else:
    tosave = np.vstack( (data.diffs[iscan],data.err[iscan] ) )
    columns = 'q diff err'.split()
    utils.saveTxt(fname,q,tosave,info=info_delay,columns=columns)

    # save one file per timedelay with all diffs for given delay
    fname = os.path.join(folder,"%sdiffs_%s.txt" % (basename,scan))
    utils.saveTxt(fname,q,data.diffs_in_scan[iscan],info=info_delay,**kw)
